"""The Airios RF bridge API entrypoint."""

from .brdg_02r13 import BRDG02R13
from .brdg_02r13 import DEFAULT_SLAVE_ID as BRDG02R13_DEFAULT_SLAVE_ID
from .client import (
    AiriosBaseTransport,
    AsyncAiriosModbusClient,
    AsyncAiriosModbusRtuClient,
    AsyncAiriosModbusTcpClient,
    AiriosRtuTransport,
    AiriosTcpTransport,
)
from .constants import BindingStatus, ProductId
from .data_model import AiriosBoundNodeInfo, AiriosData, AiriosNodeData
from .exceptions import AiriosException
from .node import AiriosNode
from .vmd_02rps78 import VMD02RPS78
from .vmn_05lm02 import VMN05LM02


class Airios:
    """The Airios RF bridge API."""

    _client: AsyncAiriosModbusClient
    bridge: BRDG02R13

    def __init__(
        self, transport: AiriosBaseTransport, slave_id: int = BRDG02R13_DEFAULT_SLAVE_ID
    ) -> None:
        """Initialize the API instance."""
        if isinstance(transport, AiriosTcpTransport):
            transport.__class__ = AiriosTcpTransport
            self._client = AsyncAiriosModbusTcpClient(transport)
        elif isinstance(transport, AiriosRtuTransport):
            transport.__class__ = AiriosRtuTransport
            self._client = AsyncAiriosModbusRtuClient(transport)
        else:
            raise AiriosException(f"Unknown trasport {transport}")
        self.bridge = BRDG02R13(slave_id, self._client)

    async def nodes(self) -> list[AiriosBoundNodeInfo]:
        """Get the list of bound nodes."""
        return await self.bridge.nodes()

    async def node(self, slave_id: int) -> AiriosNode:
        """Get a node intance by its Modbus slave ID."""
        return await self.bridge.node(slave_id)

    async def bind_status(self) -> BindingStatus:
        """Get the bind status."""
        return await self.bridge.bind_status()

    async def bind_controller(
        self,
        slave_id: int,
        product_id: ProductId,
        product_serial: int | None = None,
    ) -> bool:
        """Bind a new controller to the bridge."""
        return await self.bridge.bind_controller(slave_id, product_id, product_serial)

    async def bind_accessory(
        self,
        controller_slave_id: int,
        slave_id: int,
        product_id: ProductId,
    ) -> bool:
        """Bind a new accessory to the bridge."""
        return await self.bridge.bind_accessory(controller_slave_id, slave_id, product_id)

    async def unbind(self, slave_id: int) -> bool:
        """Remove a bound node from the bridge by its Modbus slave ID."""
        return await self.bridge.unbind(slave_id)

    async def fetch(self) -> AiriosData:
        """Get the data from all nodes at once."""
        data: dict[int, AiriosNodeData] = {}

        brdg_data = await self.bridge.fetch_bridge()
        if brdg_data["rf_address"] is None:
            raise AiriosException("Failed to fetch node RF address")
        bridge_rf_address = brdg_data["rf_address"].value
        data[self.bridge.slave_id] = brdg_data

        for node in await self.bridge.nodes():
            if node.product_id == ProductId.VMD_02RPS78:
                vmd = VMD02RPS78(node.slave_id, self.bridge.client)
                vmd_data = await vmd.fetch_vmd_data()
                data[node.slave_id] = vmd_data
            if node.product_id == ProductId.VMN_05LM02:
                vmn = VMN05LM02(node.slave_id, self.bridge.client)
                vmn_data = await vmn.fetch_vmn_data()
                data[node.slave_id] = vmn_data

        return AiriosData(bridge_rf_address=bridge_rf_address, nodes=data)

    async def connect(self) -> bool:
        """Establish underlaying Modbus connection."""
        return await self._client.connect()

    def close(self) -> None:
        """Close underlaying Modbus connection."""
        return self._client.close()



"""Airios BRDG-02R13 RF bridge implementation."""

import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import List

from .client import AsyncAiriosModbusClient
from .constants import (
    Baudrate,
    BindingMode,
    BindingStatus,
    ModbusEvents,
    Parity,
    ProductId,
    ResetMode,
    StopBits,
)
from .data_model import AiriosBoundNodeInfo, BRDG02R13Data
from .exceptions import (
    AiriosBindingException,
    AiriosException,
    AiriosInvalidArgumentException,
)
from .node import AiriosNode
from .node import Reg as NodeReg
from .registers import (
    DateTimeRegister,
    FloatRegister,
    RegisterAccess,
    RegisterAddress,
    RegisterBase,
    Result,
    StringRegister,
    U16Register,
    U32Register,
)
from .vmd_02rps78 import VMD02RPS78
from .vmn_05lm02 import VMN05LM02

DEFAULT_SLAVE_ID = 207

_LOGGER = logging.getLogger(__name__)


@dataclass
class RFLoad:
    """RF load."""

    load_current_hour: float
    """The RF load in the current hour (%)."""
    load_last_hour: float
    """The RF load in the last hour (%)."""


@dataclass
class RFSentMessages:
    """RF sent messages."""

    messages_current_hour: int
    """Messages sent in the current hour."""
    messages_last_hour: int
    """Messages sent in the last hour."""


@dataclass
class SerialConfig:
    """Serial config."""

    baudrate: Baudrate
    parity: Parity
    stop_bits: StopBits


class Reg(RegisterAddress):
    """Register set for BRDG-02R13 RF bridge."""

    CUSTOMER_PRODUCT_ID = 40023
    UTC_TIME = 41015
    LOCAL_TIME = 41017
    UPTIME = 41019
    DAYLIGHT_SAVING_TYPE = 41021
    TIMEZONE_OFFSET = 41022
    OEM_CODE = 41101
    MODBUS_EVENTS = 41103
    RESET_DEVICE = 41107
    CUSTOMER_SPECIFIC_NODE_ID = 41108
    SERIAL_PARITY = 41998
    SERIAL_STOP_BITS = 41999
    SERIAL_BAUDRATE = 42000
    SLAVE_ADDRESS = 42001
    MESSAGES_SEND_CURRENT_HOUR = 42100
    MESSAGES_SEND_LAST_HOUR = 42101
    RF_LOAD_CURRENT_HOUR = 42102
    RF_LOAD_LAST_HOUR = 42104
    BINDING_PRODUCT_ID = 43000
    BINDING_PRODUCT_SERIAL = 43002
    BINDING_COMMAND = 43004
    CREATE_NODE = 43005
    FIRST_ADDRESS_TO_ASSIGN = 43006
    REMOVE_NODE = 43399
    ACTUAL_BINDING_STATUS = 43900
    NUMBER_OF_NODES = 43901
    ADDRESS_NODE_1 = 43902
    ADDRESS_NODE_2 = 43903
    ADDRESS_NODE_3 = 43904
    ADDRESS_NODE_4 = 43905
    ADDRESS_NODE_5 = 43906
    ADDRESS_NODE_6 = 43907
    ADDRESS_NODE_7 = 43908
    ADDRESS_NODE_8 = 43909
    ADDRESS_NODE_9 = 43910
    ADDRESS_NODE_10 = 43911
    ADDRESS_NODE_11 = 43912
    ADDRESS_NODE_12 = 43913
    ADDRESS_NODE_13 = 43914
    ADDRESS_NODE_14 = 43915
    ADDRESS_NODE_15 = 43916
    ADDRESS_NODE_16 = 43917
    ADDRESS_NODE_17 = 43918
    ADDRESS_NODE_18 = 43919
    ADDRESS_NODE_19 = 43920
    ADDRESS_NODE_20 = 43921
    ADDRESS_NODE_21 = 43922
    ADDRESS_NODE_22 = 43923
    ADDRESS_NODE_23 = 43924
    ADDRESS_NODE_24 = 43925
    ADDRESS_NODE_25 = 43926
    ADDRESS_NODE_26 = 43927
    ADDRESS_NODE_27 = 43928
    ADDRESS_NODE_28 = 43929
    ADDRESS_NODE_29 = 43930
    ADDRESS_NODE_30 = 43931
    ADDRESS_NODE_31 = 43932
    ADDRESS_NODE_32 = 43933


class BRDG02R13(AiriosNode):
    """Represents a BRDG-02R13 RF bridge."""

    def __init__(self, slave_id: int, client: AsyncAiriosModbusClient) -> None:
        """Initialize the BRDG-02R13 RF bridge instance."""

        super().__init__(slave_id, client)
        brdg_registers: List[RegisterBase] = [
            U16Register(Reg.CUSTOMER_PRODUCT_ID, RegisterAccess.READ | RegisterAccess.WRITE),
            DateTimeRegister(Reg.UTC_TIME, RegisterAccess.READ | RegisterAccess.WRITE),
            DateTimeRegister(Reg.LOCAL_TIME, RegisterAccess.READ),
            U32Register(Reg.UPTIME, RegisterAccess.READ),
            U16Register(Reg.DAYLIGHT_SAVING_TYPE, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.TIMEZONE_OFFSET, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.OEM_CODE, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.MODBUS_EVENTS, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.RESET_DEVICE, RegisterAccess.WRITE),
            StringRegister(Reg.CUSTOMER_SPECIFIC_NODE_ID, 10, RegisterAccess.WRITE),
            U16Register(Reg.SERIAL_PARITY, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.SERIAL_STOP_BITS, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.SERIAL_BAUDRATE, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.SLAVE_ADDRESS, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.MESSAGES_SEND_CURRENT_HOUR, RegisterAccess.READ),
            U16Register(Reg.MESSAGES_SEND_LAST_HOUR, RegisterAccess.READ),
            FloatRegister(Reg.RF_LOAD_CURRENT_HOUR, RegisterAccess.READ),
            FloatRegister(Reg.RF_LOAD_LAST_HOUR, RegisterAccess.READ),
            U32Register(Reg.BINDING_PRODUCT_ID, RegisterAccess.READ | RegisterAccess.WRITE),
            U32Register(Reg.BINDING_PRODUCT_SERIAL, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.BINDING_COMMAND, RegisterAccess.WRITE),
            U16Register(Reg.CREATE_NODE, RegisterAccess.WRITE),
            U16Register(Reg.FIRST_ADDRESS_TO_ASSIGN, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.REMOVE_NODE, RegisterAccess.WRITE),
            U16Register(Reg.ACTUAL_BINDING_STATUS, RegisterAccess.READ),
            U16Register(Reg.NUMBER_OF_NODES, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_1, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_2, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_3, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_4, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_5, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_6, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_7, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_8, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_9, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_10, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_11, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_12, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_13, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_14, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_15, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_16, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_17, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_18, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_19, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_20, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_21, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_22, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_23, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_24, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_25, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_26, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_27, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_28, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_29, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_30, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_31, RegisterAccess.READ),
            U16Register(Reg.ADDRESS_NODE_32, RegisterAccess.READ),
        ]
        self._add_registers(brdg_registers)

    def __str__(self) -> str:
        return f"BRDG-02R13@{self.slave_id}"

    async def bind_controller(
        self,
        slave_id: int,
        product_id: ProductId,
        product_serial: int | None = None,
    ) -> bool:
        """Bind a new controller to the bridge."""
        if slave_id < 2 or slave_id > 247:
            raise AiriosInvalidArgumentException(
                f"Modbus slave address {slave_id} out of range 2-247"
            )

        if slave_id == self.slave_id:
            raise AiriosInvalidArgumentException(f"Modbus slave address {slave_id} already in use")

        mode = BindingMode.ABORT
        ok = await self.client.set_register(self.regmap[Reg.BINDING_COMMAND], mode, self.slave_id)
        if not ok:
            raise AiriosBindingException("Failed to reset binding status")

        result = await self.client.get_register(
            self.regmap[Reg.ACTUAL_BINDING_STATUS], self.slave_id
        )
        if result is None or result.value is None:
            raise AiriosBindingException("Failed to determine current binding status")
        if result.value != 0:
            raise AiriosBindingException(f"Bridge not ready for binding: {result.value}")

        mode = BindingMode.OUTGOING_SINGLE_PRODUCT
        if product_serial is not None:
            mode = BindingMode.OUTGOING_SINGLE_PRODUCT_PLUS_SERIAL

        ok = await self.client.set_register(
            self.regmap[Reg.BINDING_PRODUCT_ID], product_id, self.slave_id
        )
        if not ok:
            raise AiriosBindingException("Failed to configure binding product ID")

        ok = await self.client.set_register(self.regmap[Reg.CREATE_NODE], slave_id, self.slave_id)
        if not ok:
            raise AiriosBindingException(f"Failed to create node for slave id {slave_id}")

        if product_serial is not None:
            ok = await self.client.set_register(
                self.regmap[Reg.BINDING_PRODUCT_SERIAL], product_serial, self.slave_id
            )
            if not ok:
                raise AiriosBindingException("Failed to configure binding product serial")

        value = ((slave_id & 0xFF) << 8) | mode
        return await self.client.set_register(
            self.regmap[Reg.BINDING_COMMAND], value, self.slave_id
        )

    async def bind_status(self) -> BindingStatus:
        """Get the bind status."""
        result = await self.client.get_register(
            self.regmap[Reg.ACTUAL_BINDING_STATUS], self.slave_id
        )
        if result is None or result.value is None:
            return BindingStatus.NOT_AVAILABLE
        return BindingStatus(result.value)

    async def unbind(self, slave_id: int) -> bool:
        """Remove a bound node from the bridge by its Modbus slave ID."""
        return await self.client.set_register(self.regmap[Reg.REMOVE_NODE], slave_id, self.slave_id)

    async def bind_accessory(
        self,
        controller_slave_id: int,
        slave_id: int,
        product_id: ProductId,
    ) -> bool:
        """Bind a new accessory to the bridge."""
        if controller_slave_id < 2 or controller_slave_id > 247:
            raise AiriosInvalidArgumentException(
                f"Modbus slave address {controller_slave_id} out of range 2-247"
            )

        if slave_id < 2 or slave_id > 247:
            raise AiriosInvalidArgumentException(
                f"Modbus slave address {slave_id} out of range 2-247"
            )

        if slave_id == self.slave_id:
            raise AiriosInvalidArgumentException(f"Modbus slave address {slave_id} already in use")

        mode: BindingMode = BindingMode.ABORT
        ok = await self.client.set_register(self.regmap[Reg.BINDING_COMMAND], mode, self.slave_id)
        if not ok:
            raise AiriosBindingException("Failed to reset binding status")

        result = await self.client.get_register(
            self.regmap[Reg.ACTUAL_BINDING_STATUS], self.slave_id
        )
        if result is None or result.value is None:
            raise AiriosBindingException("Failed to determine current binding status")
        if result.value != 0:
            raise AiriosBindingException(f"Bridge not ready for binding: {result.value}")

        ok = await self.client.set_register(
            self.regmap[Reg.BINDING_PRODUCT_ID], product_id, self.slave_id
        )
        if not ok:
            raise AiriosBindingException("Failed to configure binding product ID")

        ok = await self.client.set_register(self.regmap[Reg.CREATE_NODE], slave_id, self.slave_id)
        if not ok:
            raise AiriosBindingException(f"Failed to create node for slave id {slave_id}")

        mode = BindingMode.INCOMING_ON_EXISTING_NODE
        value = ((slave_id & 0xFF) << 8) | mode
        return await self.client.set_register(
            self.regmap[Reg.BINDING_COMMAND], value, self.slave_id
        )

    async def nodes(self) -> List[AiriosBoundNodeInfo]:
        """Get the list of bound nodes."""

        reg_descs: List[RegisterBase] = [
            self.regmap[Reg.ADDRESS_NODE_1],
            self.regmap[Reg.ADDRESS_NODE_2],
            self.regmap[Reg.ADDRESS_NODE_3],
            self.regmap[Reg.ADDRESS_NODE_4],
            self.regmap[Reg.ADDRESS_NODE_5],
            self.regmap[Reg.ADDRESS_NODE_6],
            self.regmap[Reg.ADDRESS_NODE_7],
            self.regmap[Reg.ADDRESS_NODE_8],
            self.regmap[Reg.ADDRESS_NODE_9],
            self.regmap[Reg.ADDRESS_NODE_10],
            self.regmap[Reg.ADDRESS_NODE_11],
            self.regmap[Reg.ADDRESS_NODE_12],
            self.regmap[Reg.ADDRESS_NODE_13],
            self.regmap[Reg.ADDRESS_NODE_14],
            self.regmap[Reg.ADDRESS_NODE_15],
            self.regmap[Reg.ADDRESS_NODE_16],
            self.regmap[Reg.ADDRESS_NODE_17],
            self.regmap[Reg.ADDRESS_NODE_18],
            self.regmap[Reg.ADDRESS_NODE_19],
            self.regmap[Reg.ADDRESS_NODE_20],
            self.regmap[Reg.ADDRESS_NODE_21],
            self.regmap[Reg.ADDRESS_NODE_22],
            self.regmap[Reg.ADDRESS_NODE_23],
            self.regmap[Reg.ADDRESS_NODE_24],
            self.regmap[Reg.ADDRESS_NODE_25],
            self.regmap[Reg.ADDRESS_NODE_26],
            self.regmap[Reg.ADDRESS_NODE_27],
            self.regmap[Reg.ADDRESS_NODE_28],
            self.regmap[Reg.ADDRESS_NODE_29],
            self.regmap[Reg.ADDRESS_NODE_30],
            self.regmap[Reg.ADDRESS_NODE_31],
            self.regmap[Reg.ADDRESS_NODE_32],
        ]

        nodes: List[AiriosBoundNodeInfo] = []
        for item in reg_descs:
            result = await self.client.get_register(item, self.slave_id)
            if result is None or result.value is None:
                continue
            slave_id = result.value
            if slave_id == 0:
                continue

            result = await self.client.get_register(self.regmap[NodeReg.PRODUCT_ID], slave_id)
            if result is None or result.value is None:
                continue
            try:
                product_id = ProductId(result.value)
            except ValueError:
                _LOGGER.warning("Unknown product ID %s", result.value)
                continue
            else:
                product_id = ProductId(result.value)

            result = await self.client.get_register(self.regmap[NodeReg.RF_ADDRESS], slave_id)
            if result is None or result.value is None:
                continue
            rf_address = result.value

            info = AiriosBoundNodeInfo(
                slave_id=slave_id, product_id=product_id, rf_address=rf_address
            )
            nodes.append(info)
        return nodes

    async def node(self, slave_id: int) -> AiriosNode:
        """Get a node intance by its Modbus slave ID."""

        if slave_id == self.slave_id:
            return self

        for node in await self.nodes():
            if node.slave_id != slave_id:
                continue
            if node.product_id == ProductId.VMD_02RPS78:
                return VMD02RPS78(node.slave_id, self.client)
            if node.product_id == ProductId.VMN_05LM02:
                return VMN05LM02(node.slave_id, self.client)

        raise AiriosException(f"Node {slave_id} not found")

    async def rf_load_current_hour(self) -> Result[float]:
        """Get the RF load in the current hour (%)."""
        return await self.client.get_register(self.regmap[Reg.RF_LOAD_CURRENT_HOUR], self.slave_id)

    async def rf_load_last_hour(self) -> Result[float]:
        """Get the RF load in the last hour (%)."""
        return await self.client.get_register(self.regmap[Reg.RF_LOAD_LAST_HOUR], self.slave_id)

    async def rf_load(self) -> RFLoad:
        """Get the RF load."""
        r1 = await self.rf_load_current_hour()
        r2 = await self.rf_load_last_hour()
        return RFLoad(load_current_hour=r1.value, load_last_hour=r2.value)

    async def rf_sent_messages_current_hour(self) -> Result[int]:
        """Get the RF sent messages in the current hour."""
        return await self.client.get_register(
            self.regmap[Reg.MESSAGES_SEND_CURRENT_HOUR], self.slave_id
        )

    async def rf_sent_messages_last_hour(self) -> Result[int]:
        """Get the RF sent messages in the last hour."""
        return await self.client.get_register(
            self.regmap[Reg.MESSAGES_SEND_LAST_HOUR], self.slave_id
        )

    async def rf_sent_messages(self) -> RFSentMessages:
        """Get the RF sent messages."""
        r1 = await self.rf_sent_messages_current_hour()
        r2 = await self.rf_sent_messages_last_hour()
        return RFSentMessages(messages_current_hour=r1.value, messages_last_hour=r2.value)

    async def serial_config(self) -> SerialConfig:
        """Get the serial configuration."""
        result = await self.client.get_register(self.regmap[Reg.SERIAL_BAUDRATE], self.slave_id)
        baudrate: Baudrate = Baudrate(result.value)
        result = await self.client.get_register(self.regmap[Reg.SERIAL_PARITY], self.slave_id)
        parity: Parity = Parity(result.value)
        result = await self.client.get_register(self.regmap[Reg.SERIAL_STOP_BITS], self.slave_id)
        stopbits: StopBits = StopBits(result.value)
        return SerialConfig(baudrate=baudrate, stop_bits=stopbits, parity=parity)

    async def set_serial_config(self, config: SerialConfig) -> bool:
        """Set the serial configuration."""
        return (
            await self.client.set_register(
                self.regmap[Reg.SERIAL_BAUDRATE],
                config.baudrate,
                self.slave_id,
            )
            and await self.client.set_register(
                self.regmap[Reg.SERIAL_PARITY],
                config.parity,
                self.slave_id,
            )
            and await self.client.set_register(
                self.regmap[Reg.SERIAL_STOP_BITS],
                config.stop_bits,
                self.slave_id,
            )
        )

    async def modbus_events(self) -> Result[ModbusEvents]:
        """Modbus event responses via special Modbus functions."""
        result = await self.client.get_register(self.regmap[Reg.MODBUS_EVENTS], self.slave_id)
        return Result(ModbusEvents(result.value), result.status)

    async def set_modbus_events(self, value: ModbusEvents) -> bool:
        """Set Modbus event responses via special Modbus functions."""
        return await self.client.set_register(self.regmap[Reg.MODBUS_EVENTS], value, self.slave_id)

    async def power_on_time(self) -> Result[timedelta]:
        """Time since last power on or reset."""
        res = await self.client.get_register(self.regmap[Reg.UPTIME], self.slave_id)
        return Result(timedelta(seconds=res.value), res.status)

    async def reset(self, mode: ResetMode) -> bool:
        """Reset the bridge."""
        return await self.client.set_register(self.regmap[Reg.RESET_DEVICE], mode, self.slave_id)

    async def utc_time(self) -> Result[datetime]:
        """Get the UTC time."""
        return await self.client.get_register(self.regmap[Reg.UTC_TIME], self.slave_id)

    async def oem_code(self) -> Result[int]:
        """Set the bridge OEM code."""
        return await self.client.get_register(self.regmap[Reg.OEM_CODE], self.slave_id)

    async def set_oem_code(self, code: int) -> bool:
        """Set the OEM code.

        It must be set to the matching code before binding a product.
        """
        return await self.client.set_register(self.regmap[Reg.OEM_CODE], code, self.slave_id)

    async def fetch_bridge(self) -> BRDG02R13Data:  # pylint: disable=duplicate-code
        """Fetch all bridge data at once."""

        return BRDG02R13Data(
            slave_id=self.slave_id,
            rf_address=await self._safe_fetch(self.node_rf_address),
            product_id=await self._safe_fetch(self.node_product_id),
            sw_version=await self._safe_fetch(self.node_software_version),
            product_name=await self._safe_fetch(self.node_product_name),
            rf_comm_status=await self._safe_fetch(self.node_rf_comm_status),
            battery_status=await self._safe_fetch(self.node_battery_status),
            fault_status=await self._safe_fetch(self.node_fault_status),
            rf_sent_messages_last_hour=await self._safe_fetch(self.rf_sent_messages_last_hour),
            rf_sent_messages_current_hour=await self._safe_fetch(
                self.rf_sent_messages_current_hour
            ),
            rf_load_last_hour=await self._safe_fetch(self.rf_load_last_hour),
            rf_load_current_hour=await self._safe_fetch(self.rf_load_current_hour),
            power_on_time=await self._safe_fetch(self.power_on_time),
        )



"""Data model for the node data fetching functions."""

from dataclasses import dataclass
from datetime import timedelta
from typing import TypedDict

from pyairios.constants import (
    BatteryStatus,
    BoundStatus,
    FaultStatus,
    ProductId,
    RFCommStatus,
    VMDBypassMode,
    VMDBypassPosition,
    VMDErrorCode,
    VMDHeater,
    VMDRequestedVentilationSpeed,
    VMDTemperature,
    VMDVentilationSpeed,
    ValueErrorStatus,
)
from pyairios.registers import Result


@dataclass
class AiriosBoundNodeInfo:
    """Bridge bound node information."""

    slave_id: int
    product_id: ProductId
    rf_address: int


class AiriosNodeData(TypedDict):
    """Generic RF node data."""

    slave_id: int
    rf_address: Result[int] | None
    product_id: Result[ProductId] | None
    product_name: Result[str] | None
    sw_version: Result[int] | None
    rf_comm_status: Result[RFCommStatus] | None
    battery_status: Result[BatteryStatus] | None
    fault_status: Result[FaultStatus] | None


class AiriosDeviceData(AiriosNodeData):
    """Generic RF device data."""

    bound_status: Result[BoundStatus] | None
    value_error_status: Result[ValueErrorStatus] | None


class VMD02RPS78Data(AiriosDeviceData):
    """VMD-02RPS78 node data."""

    error_code: Result[VMDErrorCode] | None
    ventilation_speed: Result[VMDVentilationSpeed] | None
    override_remaining_time: Result[int] | None
    exhaust_fan_speed: Result[int] | None
    supply_fan_speed: Result[int] | None
    exhaust_fan_rpm: Result[int] | None
    supply_fan_rpm: Result[int] | None
    indoor_air_temperature: Result[VMDTemperature] | None
    outdoor_air_temperature: Result[VMDTemperature] | None
    exhaust_air_temperature: Result[VMDTemperature] | None
    supply_air_temperature: Result[VMDTemperature] | None
    filter_dirty: Result[int] | None
    filter_remaining_percent: Result[int] | None
    filter_duration_days: Result[int] | None
    bypass_position: Result[VMDBypassPosition] | None
    bypass_mode: Result[VMDBypassMode] | None
    bypass_status: Result[int] | None
    defrost: Result[int] | None
    preheater: Result[VMDHeater] | None
    postheater: Result[VMDHeater] | None
    preheater_setpoint: Result[float] | None
    free_ventilation_setpoint: Result[float] | None
    free_ventilation_cooling_offset: Result[float] | None
    frost_protection_preheater_setpoint: Result[float] | None
    preset_high_fan_speed_supply: Result[int] | None
    preset_high_fan_speed_exhaust: Result[int] | None
    preset_medium_fan_speed_supply: Result[int] | None
    preset_medium_fan_speed_exhaust: Result[int] | None
    preset_low_fan_speed_supply: Result[int] | None
    preset_low_fan_speed_exhaust: Result[int] | None
    preset_standby_fan_speed_supply: Result[int] | None
    preset_standby_fan_speed_exhaust: Result[int] | None


class BRDG02R13Data(AiriosNodeData):
    """BRDG-02R13 node data."""

    rf_sent_messages_last_hour: Result[int] | None
    rf_sent_messages_current_hour: Result[int] | None
    rf_load_last_hour: Result[float] | None
    rf_load_current_hour: Result[float] | None
    power_on_time: Result[timedelta] | None


class VMN05LM02Data(AiriosDeviceData):
    """VMN-05LM02 node data."""

    requested_ventilation_speed: Result[VMDRequestedVentilationSpeed] | None


@dataclass
class AiriosData:
    """Data from all bridge bound nodes."""

    bridge_rf_address: int
    nodes: dict[int, AiriosNodeData]



"""Async client for the Airios BRDB-02R13 Modbus gateway."""

from __future__ import annotations

import datetime
import logging
import time
import typing as t
import asyncio
from dataclasses import dataclass

import pymodbus.client as modbusClient
from pymodbus.client.mixin import ModbusClientMixin
from pymodbus.exceptions import ConnectionException as ModbusConnectionException
from pymodbus.exceptions import ModbusException, ModbusIOException
from pymodbus.pdu import ExceptionResponse, ModbusPDU
from pymodbus.pdu.register_message import (
    WriteMultipleRegistersResponse,
    WriteSingleRegisterResponse,
)

from .constants import ValueStatusFlags, ValueStatusSource
from .exceptions import (
    AiriosAcknowledgeException,
    AiriosConnectionException,
    AiriosConnectionInterruptedException,
    AiriosException,
    AiriosIOException,
    AiriosReadException,
    AiriosSlaveBusyException,
    AiriosSlaveFailureException,
    AiriosWriteException,
)
from .registers import RegisterAccess, RegisterBase, Result, ResultStatus

LOGGER = logging.getLogger(__name__)

T = t.TypeVar("T")

# Minimum time to wait between two commands sent to the device. If commands are sent too fast
# it will not respond. This happens when connected over USB.
MIN_TIME_BETWEEN_COMMANDS = 0.01


@dataclass
class AiriosBaseTransport:
    """Base class to define the bridge transport."""


@dataclass
class AiriosTcpTransport(AiriosBaseTransport):
    """Parameters for the TCP transport."""

    host: str = "192.168.0.207"
    port: int = 502


@dataclass
class AiriosRtuTransport(AiriosBaseTransport):
    """Parameters for the serial transport."""

    device: str = "/dev/ttyACM0"
    baudrate: int = 19200
    data_bits: int = 8
    parity: str = "E"
    stop_bits: int = 1


class AsyncAiriosModbusClient:
    """The base class."""

    client: modbusClient.ModbusBaseClient
    ts: float
    lock: asyncio.Lock

    def __init__(self, client: modbusClient.ModbusBaseClient) -> None:
        self.client = client
        self.ts = 0
        self.lock = asyncio.Lock()

    def __del__(self):
        if hasattr(self, "client") and self.client.connected:
            LOGGER.debug("Closing modbus connection")
            self.client.close()

    async def _reconnect(self) -> bool:
        try:
            if not self.client.connected:
                LOGGER.debug("Establishing modbus connection")
                await self.client.connect()
            if not self.client.connected:
                LOGGER.error("Failed to establish modbus connection")
                self.client.close()
                raise AiriosConnectionException
        except ModbusException as err:
            message = f"Failed to establish modbus connection: {err}"
            LOGGER.error(message)
            self.client.close()
            raise AiriosConnectionException from err
        return self.client.connected

    async def _read_registers(self, register: int, length: int, slave: int) -> ModbusPDU:
        """Async read registers from device."""

        async with self.lock:
            LOGGER.debug(
                "Reading register %s with length %s from slave %s", register, length, slave
            )

            await self._reconnect()
            try:
                elapsed = time.time() - self.ts
                if elapsed < MIN_TIME_BETWEEN_COMMANDS:
                    delay = MIN_TIME_BETWEEN_COMMANDS - elapsed
                    await asyncio.sleep(delay)

                response = await self.client.read_holding_registers(
                    register, count=length, slave=slave
                )
                if isinstance(response, ExceptionResponse):
                    if response.exception_code == ExceptionResponse.SLAVE_BUSY:
                        message = (
                            "Got a SlaveBusy Modbus Exception while reading "
                            f"register {register} (length {length}) from slave {slave}"
                        )
                        LOGGER.info(message)
                        raise AiriosSlaveBusyException(message)

                    if response.exception_code == ExceptionResponse.SLAVE_FAILURE:
                        message = (
                            "Got a SlaveFailure Modbus Exception while reading "
                            f"register {register} (length {length}) from slave {slave}"
                        )
                        LOGGER.info(message)
                        raise AiriosSlaveFailureException(message)

                    if response.exception_code == ExceptionResponse.ACKNOWLEDGE:
                        message = (
                            f"Got ACK while reading register {register} (length {length}) "
                            f"from slave {slave}."
                        )
                        LOGGER.info(message)
                        raise AiriosAcknowledgeException(message)

                    message = (
                        f"Got an error while reading register {register} "
                        f"(length {length}) from slave {slave}: {response}"
                    )
                    LOGGER.warning(message)
                    raise AiriosReadException(
                        message, modbus_exception_code=response.exception_code
                    )

                if len(response.registers) != length:
                    message = (
                        f"Mismatch between number of requested registers ({length}) "
                        f"and number of received registers ({len(response.registers)})"
                    )
                    LOGGER.error(message)
                    raise AiriosSlaveBusyException(message)
            except ModbusIOException as err:
                message = f"Could not read register, I/O exception: {err}"
                LOGGER.error(message)
                self.client.close()
                raise AiriosIOException(message) from err
            except ModbusConnectionException as err:
                message = f"Could not read register, bad connection: {err}"
                LOGGER.error(message)
                self.client.close()
                raise AiriosConnectionInterruptedException(message) from err
            except ModbusException as err:
                message = f"Modbus exception reading register: {err}"
                LOGGER.error(message)
                raise AiriosException(message) from err
            finally:
                self.ts = time.time()
            return response

    async def _write_registers(self, register: int, value: list[int], slave: int) -> bool:
        """Async write registers to device."""

        async with self.lock:
            LOGGER.debug("Writing register %s: %s to slave %s", register, value, slave)

            await self._reconnect()

            single_register = len(value) == 1
            try:
                if single_register:
                    response = await self.client.write_register(
                        register,
                        value[0],
                        slave=slave,
                    )
                else:
                    response = await self.client.write_registers(
                        register,
                        value,
                        slave=slave,
                    )
                if isinstance(response, ExceptionResponse):
                    message = (
                        f"Failed to write value {value} to register {register}: "
                        f"{response.exception_code:02X}"
                    )
                    LOGGER.info(message)
                    raise AiriosWriteException(
                        message, modbus_exception_code=response.exception_code
                    )
            except ModbusIOException as err:
                message = f"Could not write register, I/O exception: {err}"
                LOGGER.error(message)
                self.client.close()
                raise AiriosIOException(message) from err
            except ModbusConnectionException as err:
                message = f"Could not write register, bad connection: {err}"
                LOGGER.error(message)
                self.client.close()
                raise AiriosConnectionInterruptedException(message) from err
            except ModbusException as err:
                message = f"Could now write register: {err}"
                LOGGER.error(message)
                raise AiriosException(message) from err
            if single_register:
                assert isinstance(response, WriteSingleRegisterResponse)
                r1: bool = response.address == register and response.registers == [value[0]]
                return r1
            assert isinstance(response, WriteMultipleRegistersResponse)
            r2: bool = response.address == register and response.count == len(value)
            return r2

    async def get_register(self, regdesc: RegisterBase[T], slave: int) -> Result[T]:
        """Get a register from device."""

        if RegisterAccess.READ not in regdesc.description.access:
            LOGGER.warning("Attempt to read not readable register %s", regdesc)
            raise ValueError(f"Attempt to read not readable register {regdesc}")

        response = await self._read_registers(
            regdesc.description.address, regdesc.description.length, slave
        )

        value = regdesc.decode(response.registers)
        value_status = None

        if RegisterAccess.STATUS in regdesc.description.access:
            response = await self._read_registers(regdesc.description.address + 10000, 1, slave)
            tmp: int = t.cast(
                int,
                ModbusClientMixin.convert_from_registers(
                    response.registers, ModbusClientMixin.DATATYPE.UINT16, word_order="little"
                ),
            )
            if tmp is not None:
                age: int = tmp & 0x7F
                age_is_hours = (tmp >> 7) & 0x01
                flags: ValueStatusFlags = ValueStatusFlags((tmp >> 8) & 0xCF)
                source: ValueStatusSource = ValueStatusSource((tmp >> 12) & 0x03)

                if age_is_hours:
                    age *= 3600
                delta = datetime.timedelta(seconds=age)
                value_status = ResultStatus(delta, source, flags)

        return Result(value, value_status)

    async def set_register(self, register: RegisterBase[T], value: t.Any, slave: int) -> bool:
        """Write a register to the device."""

        if RegisterAccess.WRITE not in register.description.access:
            LOGGER.warning("Attempt to write not writable register %s", register)
            raise ValueError(f"Trying to write not writable register {register}")

        registers = register.encode(value)
        return await self._write_registers(register.description.address, registers, slave)

    async def connect(self) -> bool:
        """Establish underlaying Modbus connection."""
        return await self._reconnect()

    def close(self) -> None:
        """Close underlaying Modbus connection."""
        self.client.close()


class AsyncAiriosModbusTcpClient(AsyncAiriosModbusClient):
    """Airios client using Modbus TCP transport."""

    def __init__(self, transport: AiriosTcpTransport) -> None:
        client = modbusClient.AsyncModbusTcpClient(transport.host, port=transport.port)
        super().__init__(client)


class AsyncAiriosModbusRtuClient(AsyncAiriosModbusClient):
    """Airios client using Modbus RTU transport."""

    def __init__(self, transport: AiriosRtuTransport) -> None:
        client = modbusClient.AsyncModbusSerialClient(
            transport.device,
            baudrate=transport.baudrate,
            bytesize=transport.data_bits,
            parity=transport.parity,
            stopbits=transport.stop_bits,
        )
        super().__init__(client)



"""Airios VMD-02RPS78 controller implementation."""

from __future__ import annotations

import math
from dataclasses import dataclass
from typing import List

from .client import AsyncAiriosModbusClient
from .constants import (
    VMDBypassMode,
    VMDBypassPosition,
    VMDCapabilities,
    VMDErrorCode,
    VMDHeater,
    VMDHeaterStatus,
    VMDRequestedVentilationSpeed,
    VMDSensorStatus,
    VMDTemperature,
    VMDVentilationSpeed,
)
from .data_model import VMD02RPS78Data
from .device import AiriosDevice
from .exceptions import AiriosInvalidArgumentException
from .registers import (
    FloatRegister,
    RegisterAccess,
    RegisterAddress,
    RegisterBase,
    Result,
    U16Register,
)


@dataclass
class VMDPresetFansSpeeds:
    """Preset fan speeds."""

    exhaust_fan_speed: Result[int]
    """Exhaust fan speed (%)"""
    supply_fan_speed: Result[int]
    """Supply fan speed (%)"""


class Reg(RegisterAddress):
    """Register set for VMD-02RPS78 controller node."""

    CURRENT_VENTILATION_SPEED = 41000
    FAN_SPEED_EXHAUST = 41001
    FAN_SPEED_SUPPLY = 41002
    ERROR_CODE = 41003
    VENTILATION_SPEED_OVERRIDE_REMAINING_TIME = 41004
    TEMPERATURE_INDOOR = 41005
    TEMPERATURE_OUTDOOR = 41007
    TEMPERATURE_EXHAUST = 41009
    TEMPERATURE_SUPPLY = 41011
    PREHEATER = 41013
    FILTER_DIRTY = 41014
    DEFROST = 41015
    BYPASS_POSITION = 41016
    HUMIDITY_INDOOR = 41017
    HUMIDITY_OUTDOOR = 41018
    FLOW_INLET = 41019
    FLOW_OUTLET = 41021
    AIR_QUALITY = 41023
    AIR_QUALITY_BASIS = 41024
    CO2_LEVEL = 41025
    POST_HEATER = 41026
    CAPABILITIES = 41027
    FILTER_REMAINING_DAYS = 41040
    FILTER_DURATION = 41041
    FILTER_REMAINING_PERCENT = 41042
    FAN_RPM_EXHAUST = 41043
    FAN_RPM_SUPPLY = 41044
    BYPASS_MODE = 41050
    BYPASS_STATUS = 41051
    REQUESTED_VENTILATION_SPEED = 41500
    OVERRIDE_TIME_SPEED_LOW = 41501
    OVERRIDE_TIME_SPEED_MID = 41502
    OVERRIDE_TIME_SPEED_HIGH = 41503
    REQUESTED_BYPASS_MODE = 41550
    FILTER_RESET = 42000
    FAN_SPEED_AWAY_SUPPLY = 42001
    FAN_SPEED_AWAY_EXHAUST = 42002
    FAN_SPEED_LOW_SUPPLY = 42003
    FAN_SPEED_LOW_EXHAUST = 42004
    FAN_SPEED_MID_SUPPLY = 42005
    FAN_SPEED_MID_EXHAUST = 42006
    FAN_SPEED_HIGH_SUPPLY = 42007
    FAN_SPEED_HIGH_EXHAUST = 42008
    FROST_PROTECTION_PREHEATER_SETPOINT = 42009
    PREHEATER_SETPOINT = 42011
    FREE_VENTILATION_HEATING_SETPOINT = 42013
    FREE_VENTILATION_COOLING_OFFSET = 42015


class VMD02RPS78(AiriosDevice):
    """Represents a VMD-02RPS78 controller node."""

    def __init__(self, slave_id: int, client: AsyncAiriosModbusClient) -> None:
        """Initialize the VMD-02RPS78 controller node instance."""
        super().__init__(slave_id, client)
        vmd_registers: List[RegisterBase] = [
            U16Register(Reg.CURRENT_VENTILATION_SPEED, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FAN_SPEED_EXHAUST, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FAN_SPEED_SUPPLY, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.ERROR_CODE, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(
                Reg.VENTILATION_SPEED_OVERRIDE_REMAINING_TIME,
                RegisterAccess.READ | RegisterAccess.STATUS,
            ),
            FloatRegister(Reg.TEMPERATURE_INDOOR, RegisterAccess.READ | RegisterAccess.STATUS),
            FloatRegister(Reg.TEMPERATURE_OUTDOOR, RegisterAccess.READ | RegisterAccess.STATUS),
            FloatRegister(Reg.TEMPERATURE_EXHAUST, RegisterAccess.READ | RegisterAccess.STATUS),
            FloatRegister(Reg.TEMPERATURE_SUPPLY, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.PREHEATER, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FILTER_DIRTY, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.DEFROST, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.BYPASS_POSITION, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.HUMIDITY_INDOOR, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.HUMIDITY_OUTDOOR, RegisterAccess.READ | RegisterAccess.STATUS),
            FloatRegister(Reg.FLOW_INLET, RegisterAccess.READ | RegisterAccess.STATUS),
            FloatRegister(Reg.FLOW_OUTLET, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.AIR_QUALITY, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.AIR_QUALITY_BASIS, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.CO2_LEVEL, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.POST_HEATER, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.CAPABILITIES, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FILTER_REMAINING_DAYS, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FILTER_DURATION, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FILTER_REMAINING_PERCENT, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FAN_RPM_EXHAUST, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.FAN_RPM_SUPPLY, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.BYPASS_MODE, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(Reg.BYPASS_STATUS, RegisterAccess.READ | RegisterAccess.STATUS),
            U16Register(
                Reg.REQUESTED_VENTILATION_SPEED,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(Reg.OVERRIDE_TIME_SPEED_LOW, RegisterAccess.WRITE),
            U16Register(Reg.OVERRIDE_TIME_SPEED_MID, RegisterAccess.WRITE),
            U16Register(Reg.OVERRIDE_TIME_SPEED_HIGH, RegisterAccess.WRITE),
            U16Register(
                Reg.REQUESTED_BYPASS_MODE,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(Reg.FILTER_RESET, RegisterAccess.WRITE | RegisterAccess.STATUS),
            U16Register(
                Reg.FAN_SPEED_AWAY_SUPPLY,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_AWAY_EXHAUST,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_LOW_SUPPLY,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_LOW_EXHAUST,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_MID_SUPPLY,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_MID_EXHAUST,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_HIGH_SUPPLY,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            U16Register(
                Reg.FAN_SPEED_HIGH_EXHAUST,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            FloatRegister(
                Reg.FROST_PROTECTION_PREHEATER_SETPOINT,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            FloatRegister(
                Reg.PREHEATER_SETPOINT,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            FloatRegister(
                Reg.FREE_VENTILATION_HEATING_SETPOINT,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
            FloatRegister(
                Reg.FREE_VENTILATION_COOLING_OFFSET,
                RegisterAccess.READ | RegisterAccess.WRITE | RegisterAccess.STATUS,
            ),
        ]
        self._add_registers(vmd_registers)

    def __str__(self) -> str:
        return f"VMD-02RPS78@{self.slave_id}"

    async def capabilities(self) -> Result[VMDCapabilities]:
        """Get the ventilation unit capabilitires."""
        regdesc = self.regmap[Reg.CAPABILITIES]
        result = await self.client.get_register(regdesc, self.slave_id)
        return Result(VMDCapabilities(result.value), result.status)

    async def ventilation_speed(self) -> Result[VMDVentilationSpeed]:
        """Get the ventilation unit active speed preset."""
        regdesc = self.regmap[Reg.CURRENT_VENTILATION_SPEED]
        result = await self.client.get_register(regdesc, self.slave_id)
        return Result(VMDVentilationSpeed(result.value), result.status)

    async def set_ventilation_speed(self, speed: VMDRequestedVentilationSpeed) -> bool:
        """Set the ventilation unit speed preset."""
        return await self.client.set_register(
            self.regmap[Reg.REQUESTED_VENTILATION_SPEED], speed, self.slave_id
        )

    async def set_ventilation_speed_override_time(
        self, speed: VMDRequestedVentilationSpeed, minutes: int
    ) -> bool:
        """Set the ventilation unit speed preset for a limited time."""
        if minutes > 18 * 60:
            raise AiriosInvalidArgumentException("Maximum speed override time is 18 hours")
        if speed == VMDRequestedVentilationSpeed.LOW:
            return await self.client.set_register(
                self.regmap[Reg.OVERRIDE_TIME_SPEED_LOW], minutes, self.slave_id
            )
        if speed == VMDRequestedVentilationSpeed.MID:
            return await self.client.set_register(
                self.regmap[Reg.OVERRIDE_TIME_SPEED_MID], minutes, self.slave_id
            )
        if speed == VMDRequestedVentilationSpeed.HIGH:
            return await self.client.set_register(
                self.regmap[Reg.OVERRIDE_TIME_SPEED_HIGH], minutes, self.slave_id
            )
        raise AiriosInvalidArgumentException(f"Invalid temporary override speed {speed}")

    async def preset_away_fans_speed(self) -> VMDPresetFansSpeeds:
        """Get the away ventilation speed preset fan speeds."""
        r1 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_AWAY_SUPPLY], self.slave_id)
        r2 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_AWAY_EXHAUST], self.slave_id)
        return VMDPresetFansSpeeds(supply_fan_speed=r1, exhaust_fan_speed=r2)

    async def set_preset_away_fans_speed(self, supply: int, exhaust: int) -> bool:
        """Set the away ventilation speed preset fan speeds."""
        if supply < 0 or exhaust < 0:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        if supply > 40 or exhaust > 40:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        r1 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_AWAY_SUPPLY], supply, self.slave_id
        )
        r2 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_AWAY_EXHAUST], exhaust, self.slave_id
        )
        return r1 and r2

    async def preset_low_fans_speed(self) -> VMDPresetFansSpeeds:
        """Get the low ventilation speed preset fan speeds."""
        r1 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_LOW_SUPPLY], self.slave_id)
        r2 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_LOW_EXHAUST], self.slave_id)
        return VMDPresetFansSpeeds(supply_fan_speed=r1, exhaust_fan_speed=r2)

    async def set_preset_low_fans_speed(self, supply: int, exhaust: int) -> bool:
        """Set the low ventilation speed preset fan speeds."""
        if supply < 0 or exhaust < 0:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        if supply > 80 or exhaust > 80:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        r1 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_LOW_SUPPLY], supply, self.slave_id
        )
        r2 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_LOW_EXHAUST], exhaust, self.slave_id
        )
        return r1 and r2

    async def preset_mid_fans_speed(self) -> VMDPresetFansSpeeds:
        """Get the mid ventilation speed preset fan speeds."""
        r1 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_MID_SUPPLY], self.slave_id)
        r2 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_MID_EXHAUST], self.slave_id)
        return VMDPresetFansSpeeds(supply_fan_speed=r1, exhaust_fan_speed=r2)

    async def set_preset_mid_fans_speed(self, supply: int, exhaust: int) -> bool:
        """Set the mid ventilation speed preset fan speeds."""
        if supply < 0 or exhaust < 0:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        if supply > 100 or exhaust > 100:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        r1 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_MID_SUPPLY], supply, self.slave_id
        )
        r2 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_MID_EXHAUST], exhaust, self.slave_id
        )
        return r1 and r2

    async def preset_high_fans_speed(self) -> VMDPresetFansSpeeds:
        """Get the high ventilation speed preset fan speeds."""
        r1 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_HIGH_SUPPLY], self.slave_id)
        r2 = await self.client.get_register(self.regmap[Reg.FAN_SPEED_HIGH_EXHAUST], self.slave_id)
        return VMDPresetFansSpeeds(supply_fan_speed=r1, exhaust_fan_speed=r2)

    async def set_preset_high_fans_speed(self, supply: int, exhaust: int) -> bool:
        """Set the high ventilation speed preset fan speeds."""
        if supply < 0 or exhaust < 0:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        if supply > 100 or exhaust > 100:
            raise AiriosInvalidArgumentException("Speed must be in range 0-40 %")
        r1 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_HIGH_SUPPLY], supply, self.slave_id
        )
        r2 = await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_HIGH_EXHAUST], exhaust, self.slave_id
        )
        return r1 and r2

    async def bypass_mode(self) -> Result[VMDBypassMode]:
        """Get the bypass mode."""
        regdesc = self.regmap[Reg.BYPASS_MODE]
        result = await self.client.get_register(regdesc, self.slave_id)
        try:
            mode = VMDBypassMode(result.value)
        except ValueError:
            mode = VMDBypassMode.UNKNOWN
        return Result(mode, result.status)

    async def set_bypass_mode(self, mode: VMDBypassMode) -> bool:
        """Set the bypass mode."""
        if mode == VMDBypassMode.UNKNOWN:
            raise AiriosInvalidArgumentException(f"Invalid bypass mode {mode}")
        return await self.client.set_register(
            self.regmap[Reg.REQUESTED_BYPASS_MODE], mode, self.slave_id
        )

    async def bypass_status(self) -> Result[int]:
        """Get the bypass status."""
        return await self.client.get_register(self.regmap[Reg.BYPASS_STATUS], self.slave_id)

    async def bypass_position(self) -> Result[VMDBypassPosition]:
        """Get the bypass position."""
        regdesc = self.regmap[Reg.BYPASS_POSITION]
        result = await self.client.get_register(regdesc, self.slave_id)
        error = result.value > 120
        return Result(VMDBypassPosition(result.value, error), result.status)

    async def filter_duration(self) -> Result[int]:
        """Get the filter duration (in days)."""
        return await self.client.get_register(self.regmap[Reg.FILTER_DURATION], self.slave_id)

    async def filter_remaining_days(self) -> Result[int]:
        """Get the filter remaining lifetime (in days)."""
        return await self.client.get_register(self.regmap[Reg.FILTER_REMAINING_DAYS], self.slave_id)

    async def filter_remaining(self) -> Result[int]:
        """Get the filter remaining lifetime (in %)."""
        return await self.client.get_register(
            self.regmap[Reg.FILTER_REMAINING_PERCENT], self.slave_id
        )

    async def filter_reset(self) -> bool:
        """Reset the filter dirty status."""
        return await self.client.set_register(self.regmap[Reg.FILTER_RESET], 0, self.slave_id)

    async def filter_dirty(self) -> Result[int]:
        """Get the filter dirty status."""
        return await self.client.get_register(self.regmap[Reg.FILTER_DIRTY], self.slave_id)

    async def error_code(self) -> Result[VMDErrorCode]:
        """Get the ventilation unit error code."""
        regdesc = self.regmap[Reg.ERROR_CODE]
        result = await self.client.get_register(regdesc, self.slave_id)
        return Result(VMDErrorCode(result.value), result.status)

    async def exhaust_fan_speed(self) -> Result[int]:
        """Get the exhaust fan speed (%)"""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_EXHAUST], self.slave_id)

    async def supply_fan_speed(self) -> Result[int]:
        """Get the supply fan speed (%)"""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_SUPPLY], self.slave_id)

    async def exhaust_fan_rpm(self) -> Result[int]:
        """Get the exhaust fan speed (RPM)"""
        return await self.client.get_register(self.regmap[Reg.FAN_RPM_EXHAUST], self.slave_id)

    async def supply_fan_rpm(self) -> Result[int]:
        """Get the supply fan speed (RPM)"""
        return await self.client.get_register(self.regmap[Reg.FAN_RPM_SUPPLY], self.slave_id)

    async def override_remaining_time(self) -> Result[int]:
        """Get the ventilation speed override remaining time."""
        return await self.client.get_register(
            self.regmap[Reg.VENTILATION_SPEED_OVERRIDE_REMAINING_TIME], self.slave_id
        )

    async def indoor_air_temperature(self) -> Result[VMDTemperature]:
        """Get the indoor air temperature.

        This is exhaust flow before the heat exchanger.
        """
        regdesc = self.regmap[Reg.TEMPERATURE_INDOOR]
        result = await self.client.get_register(regdesc, self.slave_id)
        if math.isnan(result.value):
            status = VMDSensorStatus.UNAVAILABLE
        elif result.value < -273.0:
            status = VMDSensorStatus.ERROR
        else:
            status = VMDSensorStatus.OK
        return Result(VMDTemperature(result.value, status), result.status)

    async def outdoor_air_temperature(self) -> Result[VMDTemperature]:
        """Get the outdoor air temperature.

        This is the supply flow before the heat exchanger.
        """
        regdesc = self.regmap[Reg.TEMPERATURE_OUTDOOR]
        result = await self.client.get_register(regdesc, self.slave_id)
        if math.isnan(result.value):
            status = VMDSensorStatus.UNAVAILABLE
        elif result.value < -273.0:
            status = VMDSensorStatus.ERROR
        else:
            status = VMDSensorStatus.OK
        return Result(VMDTemperature(result.value, status), result.status)

    async def exhaust_air_temperature(self) -> Result[VMDTemperature]:
        """Get the exhaust air temperature.

        This is the exhaust flow after the heat exchanger.
        """
        regdesc = self.regmap[Reg.TEMPERATURE_EXHAUST]
        result = await self.client.get_register(regdesc, self.slave_id)
        if math.isnan(result.value):
            status = VMDSensorStatus.UNAVAILABLE
        elif result.value < -273.0:
            status = VMDSensorStatus.ERROR
        else:
            status = VMDSensorStatus.OK
        return Result(VMDTemperature(result.value, status), result.status)

    async def supply_air_temperature(self) -> Result[VMDTemperature]:
        """Get the supply air temperature.

        This is the supply flow after the heat exchanger.
        """
        regdesc = self.regmap[Reg.TEMPERATURE_SUPPLY]
        result = await self.client.get_register(regdesc, self.slave_id)
        if math.isnan(result.value):
            status = VMDSensorStatus.UNAVAILABLE
        elif result.value < -273.0:
            status = VMDSensorStatus.ERROR
        else:
            status = VMDSensorStatus.OK
        return Result(VMDTemperature(result.value, status), result.status)

    async def defrost(self) -> Result[int]:
        """Get if defrost is active."""
        return await self.client.get_register(self.regmap[Reg.DEFROST], self.slave_id)

    async def preheater(self) -> Result[VMDHeater]:
        """Get the preheater level."""
        regdesc = self.regmap[Reg.PREHEATER]
        result = await self.client.get_register(regdesc, self.slave_id)
        status = VMDHeaterStatus.UNAVAILABLE if result.value == 0xEF else VMDHeaterStatus.OK
        return Result(VMDHeater(result.value, status), result.status)

    async def postheater(self) -> Result[VMDHeater]:
        """Get the postheater level."""
        regdesc = self.regmap[Reg.POST_HEATER]
        result = await self.client.get_register(regdesc, self.slave_id)
        status = VMDHeaterStatus.UNAVAILABLE if result.value == 0xEF else VMDHeaterStatus.OK
        return Result(VMDHeater(result.value, status), result.status)

    async def preheater_setpoint(self) -> Result[float]:
        """Get the preheater setpoint."""
        return await self.client.get_register(self.regmap[Reg.PREHEATER_SETPOINT], self.slave_id)

    async def set_preheater_setpoint(self, value: float) -> bool:
        """Set the preheater setpoint."""
        return await self.client.set_register(
            self.regmap[Reg.PREHEATER_SETPOINT], value, self.slave_id
        )

    async def free_ventilation_setpoint(self) -> Result[float]:
        """Get the free ventilation setpoint."""
        return await self.client.get_register(
            self.regmap[Reg.FREE_VENTILATION_HEATING_SETPOINT], self.slave_id
        )

    async def set_free_ventilation_setpoint(self, value: float) -> bool:
        """Set the free ventilation setpoint."""
        return await self.client.set_register(
            self.regmap[Reg.FREE_VENTILATION_HEATING_SETPOINT], value, self.slave_id
        )

    async def free_ventilation_cooling_offset(self) -> Result[float]:
        """Get the free ventilation cooling offset."""
        return await self.client.get_register(
            self.regmap[Reg.FREE_VENTILATION_COOLING_OFFSET], self.slave_id
        )

    async def set_free_ventilation_cooling_offset(self, value: float) -> bool:
        """Set the free ventilation cooling offset."""
        return await self.client.set_register(
            self.regmap[Reg.FREE_VENTILATION_COOLING_OFFSET], value, self.slave_id
        )

    async def frost_protection_preheater_setpoint(self) -> Result[float]:
        """Get the frost protection preheater setpoint."""
        return await self.client.get_register(
            self.regmap[Reg.FROST_PROTECTION_PREHEATER_SETPOINT], self.slave_id
        )

    async def set_frost_protection_preheater_setpoint(self, value: float) -> bool:
        """Set the frost protection preheater setpoint."""
        return await self.client.set_register(
            self.regmap[Reg.FROST_PROTECTION_PREHEATER_SETPOINT], value, self.slave_id
        )

    async def preset_high_fan_speed_supply(self) -> Result[int]:
        """Get the supply fan speed for the high preset."""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_HIGH_SUPPLY], self.slave_id)

    async def set_preset_high_fan_speed_supply(self, value: int) -> bool:
        """Set the supply fan speed for the high preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_HIGH_SUPPLY], value, self.slave_id
        )

    async def preset_high_fan_speed_exhaust(self) -> Result[int]:
        """Get the exhaust fan speed for the high preset."""
        return await self.client.get_register(
            self.regmap[Reg.FAN_SPEED_HIGH_EXHAUST], self.slave_id
        )

    async def set_preset_high_fan_speed_exhaust(self, value: int) -> bool:
        """Set the exhaust fan speed for the high preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_HIGH_EXHAUST], value, self.slave_id
        )

    async def preset_medium_fan_speed_supply(self) -> Result[int]:
        """Get the supply fan speed for the medium preset."""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_MID_SUPPLY], self.slave_id)

    async def set_preset_medium_fan_speed_supply(self, value: int) -> bool:
        """Set the supply fan speed for the medium preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_MID_SUPPLY], value, self.slave_id
        )

    async def preset_medium_fan_speed_exhaust(self) -> Result[int]:
        """Get the exhaust fan speed for the medium preset."""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_MID_EXHAUST], self.slave_id)

    async def set_preset_medium_fan_speed_exhaust(self, value: int) -> bool:
        """Set the exhaust fan speed for the medium preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_MID_EXHAUST], value, self.slave_id
        )

    async def preset_low_fan_speed_supply(self) -> Result[int]:
        """Get the supply fan speed for the low preset."""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_LOW_SUPPLY], self.slave_id)

    async def set_preset_low_fan_speed_supply(self, value: int) -> bool:
        """Set the supply fan speed for the low preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_LOW_SUPPLY], value, self.slave_id
        )

    async def preset_low_fan_speed_exhaust(self) -> Result[int]:
        """Get the exhaust fan speed for the low preset."""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_LOW_EXHAUST], self.slave_id)

    async def set_preset_low_fan_speed_exhaust(self, value: int) -> bool:
        """Set the exhaust fan speed for the low preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_LOW_EXHAUST], value, self.slave_id
        )

    async def preset_standby_fan_speed_supply(self) -> Result[int]:
        """Get the supply fan speed for the standby preset."""
        return await self.client.get_register(self.regmap[Reg.FAN_SPEED_AWAY_SUPPLY], self.slave_id)

    async def set_preset_standby_fan_speed_supply(self, value: int) -> bool:
        """Set the supply fan speed for the standby preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_AWAY_SUPPLY], value, self.slave_id
        )

    async def preset_standby_fan_speed_exhaust(self) -> Result[int]:
        """Get the exhaust fan speed for the standby preset."""
        return await self.client.get_register(
            self.regmap[Reg.FAN_SPEED_AWAY_EXHAUST], self.slave_id
        )

    async def set_preset_standby_fan_speed_exhaust(self, value: int) -> bool:
        """Set the exhaust fan speed for the standby preset."""
        return await self.client.set_register(
            self.regmap[Reg.FAN_SPEED_AWAY_EXHAUST], value, self.slave_id
        )

    async def fetch_vmd_data(self) -> VMD02RPS78Data:  # pylint: disable=duplicate-code
        """Fetch all controller data at once."""

        return VMD02RPS78Data(
            slave_id=self.slave_id,
            rf_address=await self._safe_fetch(self.node_rf_address),
            product_id=await self._safe_fetch(self.node_product_id),
            sw_version=await self._safe_fetch(self.node_software_version),
            product_name=await self._safe_fetch(self.node_product_name),
            rf_comm_status=await self._safe_fetch(self.node_rf_comm_status),
            battery_status=await self._safe_fetch(self.node_battery_status),
            fault_status=await self._safe_fetch(self.node_fault_status),
            bound_status=await self._safe_fetch(self.device_bound_status),
            value_error_status=await self._safe_fetch(self.device_value_error_status),
            error_code=await self._safe_fetch(self.error_code),
            ventilation_speed=await self._safe_fetch(self.ventilation_speed),
            exhaust_fan_speed=await self._safe_fetch(self.exhaust_fan_speed),
            supply_fan_speed=await self._safe_fetch(self.supply_fan_speed),
            exhaust_fan_rpm=await self._safe_fetch(self.exhaust_fan_rpm),
            supply_fan_rpm=await self._safe_fetch(self.supply_fan_rpm),
            override_remaining_time=await self._safe_fetch(self.override_remaining_time),
            indoor_air_temperature=await self._safe_fetch(self.indoor_air_temperature),
            outdoor_air_temperature=await self._safe_fetch(self.outdoor_air_temperature),
            exhaust_air_temperature=await self._safe_fetch(self.exhaust_air_temperature),
            supply_air_temperature=await self._safe_fetch(self.supply_air_temperature),
            filter_dirty=await self._safe_fetch(self.filter_dirty),
            filter_remaining_percent=await self._safe_fetch(self.filter_remaining),
            filter_duration_days=await self._safe_fetch(self.filter_duration),
            defrost=await self._safe_fetch(self.defrost),
            bypass_position=await self._safe_fetch(self.bypass_position),
            bypass_mode=await self._safe_fetch(self.bypass_mode),
            bypass_status=await self._safe_fetch(self.bypass_status),
            preheater=await self._safe_fetch(self.preheater),
            postheater=await self._safe_fetch(self.postheater),
            preheater_setpoint=await self._safe_fetch(self.preheater_setpoint),
            free_ventilation_setpoint=await self._safe_fetch(self.free_ventilation_setpoint),
            free_ventilation_cooling_offset=await self._safe_fetch(
                self.free_ventilation_cooling_offset
            ),
            frost_protection_preheater_setpoint=await self._safe_fetch(
                self.frost_protection_preheater_setpoint
            ),
            preset_high_fan_speed_supply=await self._safe_fetch(self.preset_high_fan_speed_supply),
            preset_high_fan_speed_exhaust=await self._safe_fetch(
                self.preset_high_fan_speed_exhaust
            ),
            preset_medium_fan_speed_supply=await self._safe_fetch(
                self.preset_medium_fan_speed_supply
            ),
            preset_medium_fan_speed_exhaust=await self._safe_fetch(
                self.preset_medium_fan_speed_exhaust
            ),
            preset_low_fan_speed_supply=await self._safe_fetch(self.preset_low_fan_speed_supply),
            preset_low_fan_speed_exhaust=await self._safe_fetch(self.preset_low_fan_speed_exhaust),
            preset_standby_fan_speed_supply=await self._safe_fetch(
                self.preset_standby_fan_speed_supply
            ),
            preset_standby_fan_speed_exhaust=await self._safe_fetch(
                self.preset_standby_fan_speed_exhaust
            ),
        )



#!/usr/bin/env python3

"""Airios RF bridge Command Line Interface."""

import argparse
import asyncio
import logging

from aiocmd import aiocmd

try:
    from pyairios.client import AsyncAiriosModbusRtuClient
except ModuleNotFoundError:
    import os
    import sys

    sys.path.append(f"{os.path.dirname(__file__)}/src")
    from pyairios.client import AsyncAiriosModbusRtuClient

from pyairios.brdg_02r13 import (
    BRDG02R13,
    DEFAULT_SLAVE_ID as BRDG02R13_DEFAULT_SLAVE_ID,
    SerialConfig,
)
from pyairios.constants import (
    Baudrate,
    ModbusEvents,
    Parity,
    ProductId,
    ResetMode,
    StopBits,
    VMDBypassMode,
    VMDRequestedVentilationSpeed,
    VMDVentilationSpeed,
)
from pyairios.vmd_02rps78 import VMD02RPS78
from pyairios.vmn_05lm02 import VMN05LM02
from pyairios.exceptions import (
    AiriosConnectionException,
    AiriosIOException,
    AiriosInvalidArgumentException,
    AiriosNotImplemented,
)
from pyairios.client import (
    AiriosRtuTransport,
    AiriosTcpTransport,
    AsyncAiriosModbusClient,
    AsyncAiriosModbusTcpClient,
)


class AiriosVMN05LM02CLI(aiocmd.PromptToolkitCmd):
    """The VMN05LM02 CLI interface."""

    def __init__(self, vmn: VMN05LM02) -> None:
        super().__init__()
        self.prompt = f"[VMN-05LM02@{vmn.slave_id}]>> "
        self.vmn = vmn

    async def do_received_product_id(self) -> None:
        """Print the received product ID from the device."""
        res = await self.vmn.node_received_product_id()
        print(f"0x{res.value:08X}")

    async def do_requested_ventilation_speed(self) -> None:
        """Print the latest requested ventilation speed by the device."""
        res = await self.vmn.requested_ventilation_speed()
        print(f"{res}")
        if res.status is not None:
            print(f"\t{res.status}")

    async def do_status(self) -> None:
        """Print the device status."""
        res = await self.vmn.fetch_vmn_data()
        print("Node data")
        print("---------")
        print(f"    {'Product ID:': <25}{res['product_id']}")
        print(f"    {'Product Name:': <25}{res['product_name']}")
        print(f"    {'Software version:': <25}{res['sw_version']}")
        print(f"    {'RF address:': <25}{res['rf_address']}")
        print("")

        print("Device data")
        print("---------")
        print(f"    {'RF comm status:': <25}{res['rf_comm_status']}")
        print(f"    {'Battery status:': <25}{res['battery_status']}")
        print(f"    {'Fault status:': <25}{res['fault_status']}")
        print(f"    {'Bound status:': <25}{res['bound_status']}")
        print(f"    {'Value error status:': <25}{res['value_error_status']}")
        print("")

        print("VMN-02LM11 data")
        print("----------------")
        print(f"    {'Requested ventilation speed:': <40}{res['requested_ventilation_speed']}")


class AiriosVMD02RPS78CLI(aiocmd.PromptToolkitCmd):
    """The VMD02RPS78 CLI interface."""

    def __init__(self, vmd: VMD02RPS78) -> None:
        super().__init__()
        self.prompt = f"[VMD-02RPS78@{vmd.slave_id}]>> "
        self.vmd = vmd

    async def do_capabilities(self) -> None:
        """Print the device RF capabilities."""
        res = await self.vmd.capabilities()
        print(f"{res.value} ({res.status})")

    async def do_status(self) -> None:  # pylint: disable=too-many-statements
        """Print the device status."""
        res = await self.vmd.fetch_vmd_data()
        print("Node data")
        print("---------")
        print(f"    {'Product ID:': <25}{res['product_id']}")
        print(f"    {'Product Name:': <25}{res['product_name']}")
        print(f"    {'Software version:': <25}{res['sw_version']}")
        print(f"    {'RF address:': <25}{res['rf_address']}")
        print("")

        print("Device data")
        print("---------")
        print(f"    {'RF comm status:': <25}{res['rf_comm_status']}")
        print(f"    {'Battery status:': <25}{res['battery_status']}")
        print(f"    {'Fault status:': <25}{res['fault_status']}")
        print(f"    {'Bound status:': <25}{res['bound_status']}")
        print(f"    {'Value error status:': <25}{res['value_error_status']}")
        print("")

        print("VMD-02RPS78 data")
        print("----------------")
        print(f"    {'Error code:': <25}{res['error_code']}")

        print(f"    {'Ventilation speed:': <25}{res['ventilation_speed']}")
        print(f"    {'Override remaining time:': <25}{res['override_remaining_time']}")

        print(
            f"    {'Supply fan speed:': <25}{res['supply_fan_speed']}% "
            f"({res['supply_fan_rpm']} RPM)"
        )
        print(
            f"    {'Exhaust fan speed:': <25}{res['exhaust_fan_speed']}% "
            f"({res['exhaust_fan_rpm']} RPM)"
        )

        print(f"    {'Indoor temperature:': <25}{res['indoor_air_temperature']}")
        print(f"    {'Outdoor temperature:': <25}{res['outdoor_air_temperature']}")
        print(f"    {'Exhaust temperature:': <25}{res['exhaust_air_temperature']}")
        print(f"    {'Supply temperature:': <25}{res['supply_air_temperature']}")

        print(f"    {'Filter dirty:': <25}{res['filter_dirty']}")
        print(f"    {'Filter remaining:': <25}{res['filter_remaining_percent']} %")
        print(f"    {'Filter duration:': <25}{res['filter_duration_days']} days")

        print(f"    {'Bypass position:': <25}{res['bypass_position']}")
        print(f"    {'Bypass status:': <25}{res['bypass_status']}")
        print(f"    {'Bypass mode:': <25}{res['bypass_mode']}")

        print(f"    {'Defrost:': <25}{res['defrost']}")
        print(f"    {'Preheater:': <25}{res['preheater']}")
        print(f"    {'Postheater:': <25}{res['postheater']}")
        print("")

        print(f"    {'Preset speeds':<25}{'Supply':<10}{'Exhaust':<10}")
        print(f"    {'-------------':<25}")
        print(
            f"    {'High':<25}{str(res['preset_high_fan_speed_supply']) + ' %':<10}"
            f"{str(res['preset_high_fan_speed_exhaust']) + ' %':<10}"
        )
        print(
            f"    {'Mid':<25}{str(res['preset_medium_fan_speed_supply']) + ' %':<10}"
            f"{str(res['preset_medium_fan_speed_exhaust']) + ' %':<10}"
        )
        print(
            f"    {'Low':<25}{str(res['preset_low_fan_speed_supply']) + ' %':<10}"
            f"{str(res['preset_low_fan_speed_exhaust']) + ' %':<10}"
        )
        print(
            f"    {'Standby':<25}{str(res['preset_standby_fan_speed_supply']) + ' %':<10}"
            f"{str(res['preset_standby_fan_speed_exhaust']) + ' %':<10}"
        )
        print("")

        print("    Setpoints")
        print("    ---------")
        print(
            f"    {'Frost protection preheater setpoint:':<40}"
            f"{res['frost_protection_preheater_setpoint']} ºC"
        )
        print(f"    {'Preheater setpoint:': <40}{res['preheater_setpoint']} ºC")
        print(f"    {'Free ventilation setpoint:':<40}{res['free_ventilation_setpoint']} ºC")
        print(
            f"    {'Free ventilation cooling offset:':<40}"
            f"{res['free_ventilation_cooling_offset']} K"
        )

    async def do_error_code(self) -> None:
        """Print the current error code."""
        res = await self.vmd.error_code()
        print(f"{res}")

    async def do_ventilation_speed(self) -> None:
        """Print the current ventilation speed."""
        res = await self.vmd.ventilation_speed()
        if res.value in [
            VMDVentilationSpeed.OVERRIDE_LOW,
            VMDVentilationSpeed.OVERRIDE_MID,
            VMDVentilationSpeed.OVERRIDE_HIGH,
        ]:
            rem = await self.vmd.override_remaining_time()
            print(f"{res.value} ({rem.value} min. remaining)")
        else:
            print(f"{res}")
        if res.status is not None:
            print(f"{res.status}")

    async def do_set_ventilation_speed(self, preset: str) -> None:
        """Change the ventilation speed."""
        s = VMDRequestedVentilationSpeed.parse(preset)
        await self.vmd.set_ventilation_speed(s)

    async def do_set_ventilation_speed_override_time(self, preset: str, minutes: str) -> None:
        """Change the ventilation speed for a limited time."""
        s = VMDRequestedVentilationSpeed.parse(preset)
        await self.vmd.set_ventilation_speed_override_time(s, int(minutes))

    async def do_preset_away_fans_speeds(self):
        """Print the away preset fan speeds."""
        res = await self.vmd.preset_away_fans_speed()
        print(f"{'Supply fan speed:': <25}{res.supply_fan_speed}%")
        print(f"{'Exhaust fan speed:': <25}{res.exhaust_fan_speed}%")

    async def do_set_preset_away_fans_speeds(self, supply: int, exhaust: int):
        """Change the away preset fan speeds."""
        await self.vmd.set_preset_away_fans_speed(int(supply), int(exhaust))

    async def do_preset_low_fans_speeds(self):
        """Print the low preset fan speeds."""
        res = await self.vmd.preset_low_fans_speed()
        print(f"{'Supply fan speed:': <25}{res.supply_fan_speed}%")
        print(f"{'Exhaust fan speed:': <25}{res.exhaust_fan_speed}%")

    async def do_set_preset_low_fans_speeds(self, supply: int, exhaust: int):
        """Change the low preset fan speeds."""
        await self.vmd.set_preset_low_fans_speed(int(supply), int(exhaust))

    async def do_preset_mid_fans_speeds(self):
        """Print the mid preset fan speeds."""
        res = await self.vmd.preset_mid_fans_speed()
        print(f"{'Supply fan speed:': <25}{res.supply_fan_speed}%")
        print(f"{'Exhaust fan speed:': <25}{res.exhaust_fan_speed}%")

    async def do_set_preset_mid_fans_speeds(self, supply: int, exhaust: int):
        """Change the mid preset fan speeds."""
        await self.vmd.set_preset_mid_fans_speed(int(supply), int(exhaust))

    async def do_preset_high_fans_speeds(self):
        """Print the high preset fan speeds."""
        res = await self.vmd.preset_high_fans_speed()
        print(f"{'Supply fan speed:': <25}{res.supply_fan_speed}%")
        print(f"{'Exhaust fan speed:': <25}{res.exhaust_fan_speed}%")

    async def do_set_preset_high_fans_speeds(self, supply: int, exhaust: int):
        """Change the high preset fan speeds."""
        await self.vmd.set_preset_high_fans_speed(int(supply), int(exhaust))

    async def do_bypass_position(self):
        """Print the bypass position."""
        res = await self.vmd.bypass_position()
        print(f"{res}")

    async def do_bypass_status(self):
        """Print the bypass status."""
        res = await self.vmd.bypass_status()
        print(f"{res}")

    async def do_bypass_mode(self):
        """Print the bypass mode."""
        res = await self.vmd.bypass_mode()
        print(f"{res}")

    async def do_set_bypass_mode(self, mode: str):
        """Change the bypass mode."""
        v = VMDBypassMode.parse(mode)
        await self.vmd.set_bypass_mode(v)

    async def do_filter_duration(self):
        """Print the filter duration."""
        res = await self.vmd.filter_duration()
        print(f"{res}")

    async def do_filter_remaining(self):
        """Print the filter remaining percentage."""
        r1 = await self.vmd.filter_remaining()
        r2 = await self.vmd.filter_remaining_days()
        r3 = await self.vmd.filter_duration()
        print(f"{r1.value} % ({r2.value} of {r3.value} days)")

    async def do_filter_reset(self):
        """Reset the filter change timer."""
        await self.vmd.filter_reset()


class AiriosBridgeCLI(aiocmd.PromptToolkitCmd):
    """The bridge CLI interface."""

    def __init__(self, bridge: BRDG02R13) -> None:
        super().__init__()
        self.prompt = f"[BRDG-02R13@{bridge.slave_id}]>> "
        self.bridge = bridge

    async def do_nodes(self) -> None:
        """Print the list of bound nodes."""
        res = await self.bridge.nodes()
        for n in res:
            print(f"{n}")

    async def do_node(self, slave_id: str) -> None:
        """Manage a bound node."""
        nodes = await self.bridge.nodes()
        node_info = None
        for n in nodes:
            if int(slave_id) == int(n.slave_id):
                node_info = n
                break

        if node_info is None:
            raise AiriosIOException(f"Node with address {slave_id} not bound")

        if node_info.product_id == ProductId.VMD_02RPS78:
            vmd = VMD02RPS78(node_info.slave_id, self.bridge.client)
            await AiriosVMD02RPS78CLI(vmd).run()
            return

        if node_info.product_id == ProductId.VMN_05LM02:
            vmn = VMN05LM02(node_info.slave_id, self.bridge.client)
            await AiriosVMN05LM02CLI(vmn).run()
            return

        raise AiriosNotImplemented(f"{node_info.product_id} not implemented")

    async def do_rf_sent_messages(self) -> None:
        """Print the RF sent messages."""
        res = await self.bridge.rf_sent_messages()
        print(f"RF Sent Messages: {res}")

    async def do_modbus_events(self) -> None:
        """Print the modbus events mode."""
        res = await self.bridge.modbus_events()
        print(f"Modbus events: {res}")

    async def do_set_modbus_events(self, mode: str) -> None:
        """Set the Modbus events mode:
        'none'   - No Modbus events are generated
        'bridge' - Modbus function 'bridge event' is sent when a value is changed
        'node'   - Modbus function 'node event' is sent when a value is changed
        'data'   - Modbus function 'data event' is sent when a value is changed
        """
        value = ModbusEvents.parse(mode)
        await self.bridge.set_modbus_events(value)

    async def do_serial_config(self) -> None:
        """Print the serial configuration."""
        res = await self.bridge.serial_config()
        print(f"Serial Config: {res}")

    async def do_set_serial_config(self, baudrate: int, parity: str, stop_bits: int) -> None:
        """Set the serial configuration.

        The bridge must be reset to make new settings effective."""
        b = Baudrate.parse(baudrate)
        p = Parity.parse(parity)
        s = StopBits.parse(stop_bits)
        config = SerialConfig(b, p, s)
        if await self.bridge.set_serial_config(config):
            print("Reset the bridge with `reset` command to make new settings effective.")

    async def do_uptime(self) -> None:
        """Print the device uptime."""
        res = await self.bridge.power_on_time()
        print(f"Uptime: {res}")

    async def do_reset(self, factory_reset: bool = False) -> None:
        """Reset the device."""
        mode = ResetMode.SOFT_RESET
        if factory_reset:
            mode = ResetMode.FACTORY_RESET
        await self.bridge.reset(mode)

    async def do_unbind(self, slave_id) -> None:
        """Remove a bound node."""
        slave_id = int(slave_id)
        await self.bridge.unbind(slave_id)

    async def do_bind_status(self) -> None:
        """Print bind status."""
        res = await self.bridge.bind_status()
        print(f"Bind status: {res}")

    async def do_bind_controller(
        self, slave_id, product_id, product_serial: str | None = None
    ) -> None:
        """Bind a new controller."""
        slave_id = int(slave_id)
        pid = ProductId(int(product_id))
        psn = None
        if product_serial is not None:
            psn = int(product_serial)
        await self.bridge.bind_controller(slave_id, pid, psn)

    async def do_bind_accessory(self, ctrl_slave_id, slave_id, product_id) -> None:
        """Bind a new accessory."""
        ctrl_slave_id = int(ctrl_slave_id)
        slave_id = int(slave_id)
        pid = ProductId(int(product_id))
        await self.bridge.bind_accessory(ctrl_slave_id, slave_id, pid)

    async def do_software_build_date(self) -> None:
        """Print the software build date."""
        date = await self.bridge.node_software_build_date()
        print(date)

    async def do_utc_time(self) -> None:
        """Print the UTC time."""
        time = await self.bridge.utc_time()
        print(time)

    async def do_node_oem_number(self) -> None:
        """Print the node OEM number."""
        number = await self.bridge.node_oem_number()
        print(number)

    async def do_oem_code(self) -> None:
        """Print the OEM code."""
        number = await self.bridge.oem_code()
        print(number)

    async def do_set_oem_code(self, number: int) -> None:
        """Set the OEM code."""
        await self.bridge.set_oem_code(int(number))

    async def do_status(self) -> None:
        """Print the device status."""
        res = await self.bridge.fetch_bridge()
        print("Node data")
        print("---------")
        print(f"    {'Product ID:': <25}{res['product_id']}")
        print(f"    {'Product Name:': <25}{res['product_name']}")
        print(f"    {'Software version:': <25}{res['sw_version']}")
        print(f"    {'RF address:': <25}{res['rf_address']}")
        print("")

        print("Device data")
        print("---------")
        print(f"    {'RF comm status:': <25}{res['rf_comm_status']}")
        print(f"    {'Battery status:': <25}{res['battery_status']}")
        print(f"    {'Fault status:': <25}{res['fault_status']}")
        print("")

        print("BRDG-02R13 data")
        print("----------------")
        print(f"    {'RF sent messages last hour': <40}{res['rf_sent_messages_last_hour']}")
        print(f"    {'RF sent messages current hour:': <40}{res['rf_sent_messages_current_hour']}")
        print(f"    {'RF load last hour:': <40}{res['rf_load_last_hour']}")
        print(f"    {'RF load current hour:': <40}{res['rf_load_current_hour']}")
        print(f"    {'Uptime:': <40}{res['power_on_time']}")


class AiriosClientCLI(aiocmd.PromptToolkitCmd):  # pylint: disable=too-few-public-methods
    """CLI client interface."""

    prompt = "[client]>> "

    def __init__(self, client: AsyncAiriosModbusClient) -> None:
        super().__init__()
        self.client = client

    async def do_bridge(self, address: str | None = None) -> None:
        """Manage the bridge."""
        if address is None:
            _address = (
                BRDG02R13_DEFAULT_SLAVE_ID
                if isinstance(self.client, AsyncAiriosModbusRtuClient)
                else 1
            )
        else:
            _address = int(address)
        bridge = BRDG02R13(_address, self.client)
        await AiriosBridgeCLI(bridge).run()


class AiriosRootCLI(aiocmd.PromptToolkitCmd):
    """CLI root context."""

    prompt = ">> "
    intro = 'Welcome to AiriosCLI. Type "help" for available commands.'
    client: AsyncAiriosModbusClient | None = None

    async def do_connect_rtu(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        port: str = "/dev/ttyACM0",
        baudrate: str = "19200",
        data_bits: str = "8",
        parity: str = "E",
        stop_bits: str = "1",
    ) -> None:
        """Connect to serial bridge."""
        if self.client:
            raise AiriosConnectionException("Already connected")
        transport = AiriosRtuTransport(
            device=port,
            baudrate=int(baudrate),
            data_bits=int(data_bits),
            parity=parity,
            stop_bits=int(stop_bits),
        )
        self.client = AsyncAiriosModbusRtuClient(transport)
        await AiriosClientCLI(self.client).run()

    async def do_connect_tcp(self, host: str = "192.168.1.254", port: int = 502):
        """Connect to Ethernet bridge."""
        if self.client:
            raise AiriosConnectionException("Already connected")
        transport = AiriosTcpTransport(host, port=port)
        self.client = AsyncAiriosModbusTcpClient(transport)
        await AiriosClientCLI(self.client).run()

    async def do_disconnect(self) -> None:
        """Disconnect from bridge."""
        if self.client:
            self.client = None

    async def do_set_log_level(self, level: str) -> None:
        "Set the log level: critical, fatal, error, warning, info or debug."
        logging.basicConfig()
        log = logging.getLogger()
        if level.casefold() == "critical".casefold():
            log.setLevel(logging.CRITICAL)
        elif level.casefold() == "fatal".casefold():
            log.setLevel(logging.FATAL)
        elif level.casefold() == "error".casefold():
            log.setLevel(logging.ERROR)
        elif level.casefold() == "warning".casefold():
            log.setLevel(logging.WARNING)
        elif level.casefold() == "info".casefold():
            log.setLevel(logging.INFO)
        elif level.casefold() == "debug".casefold():
            log.setLevel(logging.DEBUG)
        else:
            raise AiriosInvalidArgumentException("Invalid log level")


async def main() -> None:
    """Run the async CLI."""
    await AiriosRootCLI().run()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Airios RF bridge Command Line Interface",
        epilog=(
            "Thanks to Siber for providing the documentation and suppport to develop this library."
        ),
    )
    args = parser.parse_args()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(main())



"""Constants and datay types used by this library."""

from dataclasses import dataclass
from enum import Flag, IntEnum, auto
import datetime


class ProductId(IntEnum):
    """The product ID is a unique product identifier.

    The value is composed by three fields, poduct type + sub ID + manufacturer ID.
    """

    BRDG_02R13 = 0x0001C849
    VMD_02RPS78 = 0x0001C892
    VMN_05LM02 = 0x0001C83E
    VMN_02LM11 = 0x0001C852

    def __str__(self) -> str:
        if self.value == self.BRDG_02R13:
            return "BRDG-02R13"
        if self.value == self.VMD_02RPS78:
            return "VMD-02RPS78"
        if self.value == self.VMN_05LM02:
            return "VMN-05LM02"
        if self.value == self.VMN_02LM11:
            return "VMN-02LM11"
        raise ValueError(f"Unknown product ID value {self.value}")


class BoundStatus(IntEnum):
    """RF device bound status."""

    NO_CHANGE = 0
    """No change in bound status."""
    REBOUND = 1
    """Device is re-bound to the same controller."""
    NEW_BOUND = 2
    """Device is bound for the first time to the controller."""

    def __str__(self) -> str:
        if self.value == self.NO_CHANGE:
            return "no_change"
        if self.value == self.REBOUND:
            return "rebound"
        if self.value == self.NEW_BOUND:
            return "new_bound"
        raise ValueError(f"Unknown bound status value {self.value}")


class RFCommStatus(IntEnum):
    """Node RF Communication status."""

    NO_ERROR = 0
    """No error."""
    ERROR = 1
    """No data received for 30 minutes."""

    def __str__(self) -> str:
        if self.value == self.NO_ERROR:
            return "no_error"
        if self.value == self.ERROR:
            return "error"
        raise ValueError(f"Unknown RF comm status {self.value}")


@dataclass
class BatteryStatus:
    """Node battery status."""

    available: bool
    """False means unknown or no battery present."""
    low: bool
    """True if battery is low. Meaningfull only if available is true."""


@dataclass
class FaultStatus:
    """Node fault status."""

    available: bool
    """False means unknown or device does not support faults."""
    fault: bool
    """True if faults are active. Meaningfull only if available is true."""


class ValueErrorStatus(IntEnum):
    """RF device value error status.

    This is when a value is out of range due to a broken sensor for example.
    """

    NO_ERROR = 0
    """No active value errors."""
    ERROR = 1
    """One or more value errors are active."""

    def __str__(self) -> str:
        if self.value == self.NO_ERROR:
            return "no_error"
        if self.value == self.ERROR:
            return "error"
        raise ValueError(f"Unknown value error status {self.value}")


@dataclass
class RFStats:
    """RF node statistics."""

    @dataclass
    class Record:  # pylint: disable=too-many-instance-attributes
        """RF statistic record."""

        device_id: int
        averate: int
        """Average received signal strength margin  of RF beacon (dB)."""
        stddev: float
        """Standard deviation of received signal strength margin of RF beacon (.1 dB)."""
        minimum: int
        """Lowest received signal strength (dB)."""
        maximum: int
        """Maximum received signal strength (dB)."""
        missed: int
        """Missed messages (%)."""
        received: int
        """Received beacons counter."""
        age: datetime.timedelta
        """Time since last beacon."""

    records: list[Record]


class BindingMode(IntEnum):
    """Binding mode."""

    OUTGOING_SINGLE_PRODUCT = 0x0003
    """Binds a controller."""
    OUTGOING_SINGLE_PRODUCT_PLUS_SERIAL = 0x0004
    """Binds a controller by RF address (serial number)."""
    INCOMING_ON_EXISTING_NODE = 0x0014
    """Abort an ongoing bind."""
    ABORT = 0x00C8


class BindingStatus(IntEnum):
    """Bind result."""

    NOT_AVAILABLE = 0
    """No binding initialized."""
    OUTGOING_BINDING_INITIALIZED = 1
    """Outgoing binding in process."""
    OUTGOING_BINDING_COMPLETED = 2
    """Outgoing binding completed."""
    INCOMING_BINDING_ACTIVE = 3
    """Incoming binding in process."""
    INCOMING_BINDING_COMPLETED = 4
    """Incoming binding completed."""
    LEARNING_COMPLETED = 5
    """Learning completed."""
    INCOMING_AUTODETECT_WINDOW_CLOSED = 10
    """Incoming binding autodection window closed."""
    OUTGOING_BINDING_FAILED_NO_ANSWER = 100
    """Outgoing binding failed, no answer."""
    OUTGOING_BINDING_FAILED_INCOMPATIBLE_DEVICE = 101
    """Outgoing binding failed, incompatible device."""
    OUTGOING_BINDING_FAILED_NODE_LIST_FULL = 102
    """Outgoing binding failed, node list full."""
    OUTGOING_BINDING_FAILED_MODBUS_ADDR_INVALID = 103
    """Outgoing binding failed, modbus address invalid."""
    INCOMING_BINDING_WINDOW_CLOSED_WITHOUT_BINDING_A_PRODUCT = 104
    """Incoming binding failed, window closed without binding a product."""
    BINDING_FAILED_SERIAL_NUMBER_INVALID = 105
    """Binding failed, invalid serial number."""
    UNKNOWN_BINDING_COMMAND = 200
    """Binding failed, unknown binding command."""
    UNKNOWN_PRODUCT_TYPE = 201
    """Binding failed, unknown product type."""

    def __str__(self) -> str:  # pylint: disable=too-many-return-statements,too-many-branches
        if self.value == self.NOT_AVAILABLE:
            return "not_available"
        if self.value == self.OUTGOING_BINDING_INITIALIZED:
            return "outgoing_binding_initialized"
        if self.value == self.OUTGOING_BINDING_COMPLETED:
            return "outgoing_binding_completed"
        if self.value == self.INCOMING_BINDING_ACTIVE:
            return "incoming_binding_active"
        if self.value == self.INCOMING_BINDING_COMPLETED:
            return "incoming_binding_completed"
        if self.value == self.LEARNING_COMPLETED:
            return "learning_completed"
        if self.value == self.INCOMING_AUTODETECT_WINDOW_CLOSED:
            return "incoming_autodetect_window_closed"
        if self.value == self.OUTGOING_BINDING_FAILED_NO_ANSWER:
            return "outgoing_binding_failed_no_answer"
        if self.value == self.OUTGOING_BINDING_FAILED_INCOMPATIBLE_DEVICE:
            return "outgoing_binding_failed_incompatible_device"
        if self.value == self.OUTGOING_BINDING_FAILED_NODE_LIST_FULL:
            return "outgoing_binding_failed_no_list_full"
        if self.value == self.OUTGOING_BINDING_FAILED_MODBUS_ADDR_INVALID:
            return "outgoing_binding_failed_modbus_address_invalid"
        if self.value == self.INCOMING_BINDING_WINDOW_CLOSED_WITHOUT_BINDING_A_PRODUCT:
            return "incoming_binding_window_closed_without_binding_a_product"
        if self.value == self.BINDING_FAILED_SERIAL_NUMBER_INVALID:
            return "binding_failed_serial_number_invalid"
        if self.value == self.UNKNOWN_BINDING_COMMAND:
            return "unknown_binding_command"
        if self.value == self.UNKNOWN_PRODUCT_TYPE:
            return "unknown_product_type"
        raise ValueError(f"Unknown binding status value {self.value}")


class ModbusEvents(IntEnum):
    """Modbus events enumeration."""

    NO_EVENTS = 0
    """No modbus events are generated."""
    BRIDGE_EVENTS = 1
    """Modbus function 'bridge event' is sent to master when a value is changed."""
    NODE_EVENTS = 2
    """Modbus function 'node event' is sent to master when a value is changed."""
    DATA_EVENTS = 3
    """Modbus function 'data event' is sent to master when a value is changed."""

    def __str__(self):
        if self.value == self.NO_EVENTS:
            return "no_events"
        if self.value == self.BRIDGE_EVENTS:
            return "bridge_events"
        if self.value == self.NODE_EVENTS:
            return "node_events"
        if self.value == self.DATA_EVENTS:
            return "data_events"
        raise ValueError(f"Unknown modbus event value {self.value}")

    @classmethod
    def parse(cls, value: str):  # pylint: disable=too-many-return-statements
        """Instantiate by string."""
        if value.casefold() == "none".casefold():
            return cls(cls.NO_EVENTS)
        if value.casefold() == "bridge".casefold():
            return cls(cls.BRIDGE_EVENTS)
        if value.casefold() == "node".casefold():
            return cls(cls.NODE_EVENTS)
        if value.casefold() == "data".casefold():
            return cls(cls.DATA_EVENTS)
        raise ValueError(f"Unknown modbus_events value {value}")


class ResetMode(IntEnum):
    """Reset modes."""

    SOFT_RESET = 12345
    """Soft reset."""
    FACTORY_RESET = 56789
    """Factory reset"""


class Baudrate(IntEnum):
    """Serial port baudrate."""

    BAUD_300 = 0
    BAUD_600 = 1
    BAUD_1200 = 2
    BAUD_2400 = 3
    BAUD_4800 = 4
    BAUD_9600 = 5
    BAUD_19200 = 6
    BAUD_38400 = 7
    BAUD_57600 = 8
    BAUD_115200 = 9

    @classmethod
    def parse(cls, value: int | str):  # pylint: disable=too-many-return-statements
        """Instantiate by string."""
        if int(value) == 300:
            return cls(cls.BAUD_300)
        if int(value) == 600:
            return cls(cls.BAUD_600)
        if int(value) == 1200:
            return cls(cls.BAUD_1200)
        if int(value) == 2400:
            return cls(cls.BAUD_2400)
        if int(value) == 4800:
            return cls(cls.BAUD_4800)
        if int(value) == 9600:
            return cls(cls.BAUD_9600)
        if int(value) == 19200:
            return cls(cls.BAUD_19200)
        if int(value) == 38400:
            return cls(cls.BAUD_38400)
        if int(value) == 57600:
            return cls(cls.BAUD_57600)
        if int(value) == 115200:
            return cls(cls.BAUD_115200)
        raise ValueError(f"Unknown baudrate value {value}")


class Parity(IntEnum):
    """Serial port parity."""

    PARITY_NONE = 0
    PARITY_ODD = 1
    PARITY_EVEN = 2

    @classmethod
    def parse(cls, value: str):
        """Instantiate by string."""
        if value.casefold() == "none".casefold() or value.casefold() == "N".casefold():
            return cls(cls.PARITY_NONE)
        if value.casefold() == "odd".casefold() or value.casefold() == "O".casefold():
            return cls(cls.PARITY_ODD)
        if value.casefold() == "even".casefold() or value.casefold() == "E".casefold():
            return cls(cls.PARITY_EVEN)
        raise ValueError(f"Unknown parity value {value}")


class StopBits(IntEnum):
    """Serial port stop bits."""

    STOP_1 = 0
    STOP_2 = 1

    @classmethod
    def parse(cls, value: int | str):
        """Instantiate by string."""
        if int(value) == 1:
            return cls(cls.STOP_1)
        if int(value) == 2:
            return cls(cls.STOP_2)
        raise ValueError(f"Unknown stop_bits value {value}")


class VMDSensorStatus(IntEnum):
    """VMD sensor status."""

    UNAVAILABLE = auto()
    OK = auto()
    ERROR = auto()


class VMDHeaterStatus(IntEnum):
    """VMD heater status."""

    OK = 1
    UNAVAILABLE = 2


@dataclass
class VMDTemperature:
    """VMD temperature sample."""

    temperature: float
    status: VMDSensorStatus


@dataclass
class VMDHeater:
    """VMD heater state sample."""

    level: int
    status: VMDHeaterStatus


class VMDCapabilities(Flag):
    """Ventilation unit capabilities."""

    PRE_HEATER_AVAILABLE = 0x0001
    POST_HEATER_AVAILABLE = 0x0002
    RESERVED = 0x0004
    NIGHT_MODE_CAPABLE = 0x0008
    SPEED_10_CAPABLE = 0x0010
    SPEED_9_CAPABLE = 0x0020
    SPEED_8_CAPABLE = 0x0040
    SPEED_7_CAPABLE = 0x0080
    SPEED_6_CAPABLE = 0x0100
    SPEED_5_CAPABLE = 0x0200
    SPEED_4_CAPABLE = 0x0400
    AUTO_MODE_CAPABLE = 0x0800
    BOOST_MODE_CAPABLE = 0x1000
    TIMER_CAPABLE = 0x2000
    AWAY_MODE_CAPABLE = 0x4000
    OFF_CAPABLE = 0x8000


class VMDFaultStatus(IntEnum):
    """VMD fault status codes."""

    OK = 0
    FAN_FAILURE = 1


class VMDVentilationSpeed(IntEnum):
    """Ventilation unit speed preset."""

    OFF = 0
    LOW = 1
    MID = 2
    HIGH = 3
    OVERRIDE_LOW = 11
    OVERRIDE_MID = 12
    OVERRIDE_HIGH = 13
    AWAY = 21
    BOOST = 23
    AUTO = 24

    def __str__(self) -> str:  # pylint: disable=too-many-return-statements
        if self.value == self.OFF:
            return "Off"
        if self.value == self.LOW:
            return "Low"
        if self.value == self.MID:
            return "Mid"
        if self.value == self.HIGH:
            return "High"
        if self.value == self.OVERRIDE_LOW:
            return "Low (temporary override)"
        if self.value == self.OVERRIDE_MID:
            return "Mid (temporary override)"
        if self.value == self.OVERRIDE_HIGH:
            return "High (temporary override)"
        if self.value == self.AWAY:
            return "Away"
        if self.value == self.BOOST:
            return "Boost"
        if self.value == self.AUTO:
            return "Auto"
        raise ValueError(f"Unknown ventilation speed value {self.value}")


class VMDRequestedVentilationSpeed(IntEnum):
    """VMD Requested ventilation speed codes."""

    OFF = 0
    AWAY = 1
    LOW = 2
    MID = 3
    HIGH = 4
    AUTO = 5
    BOOST = 7

    def __str__(self) -> str:  # pylint: disable=too-many-return-statements
        if self.value == self.OFF:
            return "Off"
        if self.value == self.LOW:
            return "Low"
        if self.value == self.MID:
            return "Mid"
        if self.value == self.HIGH:
            return "High"
        if self.value == self.AWAY:
            return "Away"
        if self.value == self.BOOST:
            return "Boost"
        if self.value == self.AUTO:
            return "Auto"
        raise ValueError(f"Unknown requested ventilation speed value {self.value}")

    @classmethod
    def parse(cls, value: str):  # pylint: disable=too-many-return-statements
        """Instantiate by string."""
        if value.casefold() == "off".casefold():
            return cls(cls.OFF)
        if value.casefold() == "low".casefold():
            return cls(cls.LOW)
        if value.casefold() == "mid".casefold():
            return cls(cls.MID)
        if value.casefold() == "high".casefold():
            return cls(cls.HIGH)
        if value.casefold() == "away".casefold():
            return cls(cls.AWAY)
        if value.casefold() == "boost".casefold():
            return cls(cls.BOOST)
        if value.casefold() == "auto".casefold():
            return cls(cls.AUTO)
        raise ValueError(f"Unknown requested ventilation speed value {value}")


class ValueStatusFlags(Flag):
    """Register value status flags."""

    VALID = 0x01
    """Data is valid. Invalid means never written/received."""
    ERROR = 0x02
    """Value has an error, for example a sensor error."""
    READ_PENDING = 0x04
    """Value has a read pending."""
    WRITE_PENDING = 0x08
    """Value has a write pending."""
    NEW_VALUE = 0x40
    """There is a new value cached in the bridge."""


class ValueStatusSource(IntEnum):
    """Register value source."""

    UNKNOWN = 0
    """Unknown, never written or received."""
    RF = 1
    """Value is from RF."""
    MODBUS = 2
    """Value is from Modbus interface."""

    def __str__(self) -> str:
        if self.value == self.UNKNOWN:
            return "Unknown"
        if self.value == self.RF:
            return "RF"
        if self.value == self.MODBUS:
            return "Modbus"
        raise ValueError(f"Unknown value source value {self.value}")


class VMDErrorCode(IntEnum):
    """Ventilation unit error codes."""

    NO_ERROR = 0
    NON_SPECIFIC_FAULT = 1
    EMERGENCY_STOP = 2
    FAN_1_ERROR = 3
    X22_SENSOR_ERROR = 4
    X23_SENSOR_ERROR = 5
    X21_SENSOR_ERROR = 6
    X20_SENSOR_ERROR = 7
    FAN_2_ERROR = 8
    BINDING_MODE_ACTIVE = 254
    IDENTIFICATION_ACTIVE = 255

    def __str__(self) -> str:  # pylint: disable=too-many-return-statements
        if self.value == self.NO_ERROR:
            return "no_error"
        if self.value == self.NON_SPECIFIC_FAULT:
            return "non_specific_fault"
        if self.value == self.EMERGENCY_STOP:
            return "emergency_stop"
        if self.value == self.FAN_1_ERROR:
            return "fan_1_error"
        if self.value == self.X22_SENSOR_ERROR:
            return "x22_sensor_error"
        if self.value == self.X23_SENSOR_ERROR:
            return "x23_sensor_error"
        if self.value == self.X21_SENSOR_ERROR:
            return "x21_sensor_error"
        if self.value == self.X20_SENSOR_ERROR:
            return "x20_sensor_error"
        if self.value == self.FAN_2_ERROR:
            return "fan_2_error"
        if self.value == self.BINDING_MODE_ACTIVE:
            return "binding_mode_active"
        if self.value == self.IDENTIFICATION_ACTIVE:
            return "identification_active"
        raise ValueError(f"Unknown error code {self.value}")


class VMDBypassMode(IntEnum):
    """VMD bypass mode codes."""

    CLOSE = 0
    OPEN = 100
    UNKNOWN = 239
    AUTO = 255

    def __str__(self) -> str:
        if self.value == self.CLOSE:
            return "closed"
        if self.value == self.OPEN:
            return "open"
        if self.value == self.UNKNOWN:
            return "unknown"
        if self.value == self.AUTO:
            return "auto"
        raise ValueError(f"Unknown bypass mode {self.value}")

    @classmethod
    def parse(cls, value: str):
        """Instantiate by string."""
        if value.casefold() == "close".casefold():
            return cls(cls.CLOSE)
        if value.casefold() == "open".casefold():
            return cls(cls.OPEN)
        if value.casefold() == "auto".casefold():
            return cls(cls.AUTO)
        raise ValueError(f"Unknown bypass mode {value}")


@dataclass
class VMDBypassPosition:
    """VMD bypass position sample."""

    position: int
    error: bool



"""Register definitions."""

import datetime
import struct
import typing as t
from dataclasses import dataclass
from enum import Flag, IntEnum, auto

from pymodbus.client.mixin import ModbusClientMixin

from .constants import ValueStatusFlags, ValueStatusSource
from .exceptions import AiriosDecodeError, AiriosInvalidArgumentException

T = t.TypeVar("T")


class RegisterAddress(IntEnum):
    """The register address base class."""


class RegisterAccess(Flag):
    """Register access flags."""

    READ = auto()
    WRITE = auto()
    STATUS = auto()


@dataclass(frozen=True)
class RegisterDescription:
    """Register description."""

    address: RegisterAddress
    length: int
    access: RegisterAccess


class RegisterBase(t.Generic[T]):
    """Base class for register definitions."""

    description: RegisterDescription
    datatype: ModbusClientMixin.DATATYPE

    def __init__(self, description: RegisterDescription) -> None:
        """Initialize the register instance."""
        self.description = description

    def decode(self, registers: list[int]) -> T:
        """Decode register bytes to value."""
        return ModbusClientMixin.convert_from_registers(
            registers, self.datatype, word_order="little"
        )  # type: ignore

    def encode(self, value: T) -> list[int]:
        """Encode value to register bytes."""
        return ModbusClientMixin.convert_to_registers(
            value,  # type: ignore
            self.datatype,
            word_order="little",
        )


class StringRegister(RegisterBase[str]):
    """String register."""

    datatype = ModbusClientMixin.DATATYPE.STRING

    def __init__(self, address: RegisterAddress, length: int, access: RegisterAccess) -> None:
        """Initialize the StringRegister instance."""
        description = RegisterDescription(address, length, access)
        super().__init__(description)

    def decode(self, registers: list[int]) -> str:
        """Decode register bytes to value."""

        def registers_to_bytearray(registers: list[int]) -> bytearray:
            """Convert registers to bytes."""
            b = bytearray()
            for x in registers:
                b.extend(x.to_bytes(2, "big"))
            return b

        b = registers_to_bytearray(registers)

        # remove trailing null bytes
        trailing_nulls_begin = len(b)
        while trailing_nulls_begin > 0 and b[trailing_nulls_begin - 1] == 0:
            trailing_nulls_begin -= 1

        b = b[:trailing_nulls_begin]

        try:
            result = b.decode("utf-8")
        except UnicodeDecodeError as err:
            raise AiriosDecodeError from err
        return result

    def encode(self, value: str) -> list[int]:
        return ModbusClientMixin.convert_to_registers(value, self.datatype, word_order="little")


class NumberRegister(RegisterBase[T]):
    """Base class for number registers."""

    def decode(self, registers: list[int]) -> T:
        """Decode register bytes to value."""
        result: T = t.cast(
            T,
            ModbusClientMixin.convert_from_registers(registers, self.datatype, word_order="little"),
        )
        return result

    def encode(self, value: T) -> list[int]:
        """Encode value to register bytes."""
        if isinstance(value, int):
            int_value = value
        elif isinstance(value, float):
            int_value = int(value)
        elif isinstance(value, bool):
            int_value = int(value)
        else:
            raise AiriosInvalidArgumentException(f"Unsupported type {type(value)}")
        return ModbusClientMixin.convert_to_registers(int_value, self.datatype, word_order="little")


class U16Register(NumberRegister[int]):
    """Unsigned 16-bit register."""

    datatype = ModbusClientMixin.DATATYPE.UINT16

    def __init__(self, address: RegisterAddress, access: RegisterAccess) -> None:
        """Initialize the U16Register instance."""
        description = RegisterDescription(address, 1, access)
        super().__init__(description)


class I16Register(NumberRegister[int]):
    """Signed 16-bit register."""

    datatype = ModbusClientMixin.DATATYPE.INT16

    def __init__(self, address: RegisterAddress, access: RegisterAccess) -> None:
        """Initialize the I16Register instance."""
        description = RegisterDescription(address, 1, access)
        super().__init__(description)


class U32Register(NumberRegister[int]):
    """Unsigned 32-bit register."""

    datatype = ModbusClientMixin.DATATYPE.UINT32

    def __init__(self, address: RegisterAddress, access: RegisterAccess) -> None:
        """Initialize the U32Register instance."""
        description = RegisterDescription(address, 2, access)
        super().__init__(description)


class FloatRegister(NumberRegister[float]):
    """Float register."""

    datatype = ModbusClientMixin.DATATYPE.FLOAT32

    def __init__(self, address: RegisterAddress, access: RegisterAccess) -> None:
        """Initialize the FloatRegister instance."""
        description = RegisterDescription(address, 2, access)
        super().__init__(description)


class DateRegister(RegisterBase[datetime.date]):
    """Date register."""

    datatype = ModbusClientMixin.DATATYPE.UINT32

    def __init__(self, address: RegisterAddress, access: RegisterAccess) -> None:
        """Initialize the DateRegister instance."""
        description = RegisterDescription(address, 2, access)
        super().__init__(description)

    def decode(self, registers: list[int]) -> datetime.date:
        """Decode register bytes to value."""
        value: int = t.cast(int, super().decode(registers))

        if value == 0xFFFFFFFF:
            return datetime.date.min

        buf = value.to_bytes(4, "big")
        (day, month, year) = struct.unpack(">BBH", buf)
        return datetime.date(year, month, day)

    def encode(self, value: datetime.date) -> list[int]:
        """Encode value to register bytes."""

        buf = struct.pack(">BBH", value.day, value.month, value.year)
        ival = int.from_bytes(buf, byteorder="big")
        return ModbusClientMixin.convert_to_registers(ival, self.datatype, word_order="little")


class DateTimeRegister(RegisterBase[datetime.datetime]):
    """DateTime register."""

    datatype = ModbusClientMixin.DATATYPE.UINT32

    def __init__(self, address: RegisterAddress, access: RegisterAccess) -> None:
        """Initialize the DateTimeRegister instance."""
        description = RegisterDescription(address, 2, access)
        super().__init__(description)

    def decode(self, registers: list[int]) -> datetime.datetime:
        """Decode register bytes to value."""
        value: int = t.cast(int, super().decode(registers))

        if value == 0xFFFFFFFF:
            return datetime.datetime.min

        return datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)

    def encode(self, value: datetime.datetime) -> list[int]:
        """Encode value to register bytes."""

        ts: float = value.replace(tzinfo=datetime.timezone.utc).timestamp()
        ival: int = int(ts)
        return ModbusClientMixin.convert_to_registers(ival, self.datatype, word_order="little")


@dataclass
class ResultStatus:
    """Metadata associated to a register value."""

    age: datetime.timedelta
    source: ValueStatusSource
    flags: ValueStatusFlags

    def __str__(self) -> str:
        return f"value is {self.age} old, last seen from {self.source}, flags: {self.flags}"


@dataclass
class Result(t.Generic[T]):
    """Register read result."""

    value: T
    status: ResultStatus | None

    def __init__(self, value: T, status: ResultStatus | None = None) -> None:
        super().__init__()
        self.value = value
        self.status = status

    def __str__(self) -> str:
        if isinstance(self.value, str):
            return f"{self.value}"
        if isinstance(self.value, datetime.date):
            return f"{self.value.isoformat()}"
        if isinstance(self.value, datetime.datetime):
            return f"{self.value.isoformat()}"
        if isinstance(self.value, float):
            return f"{self.value:.04f}"
        if isinstance(self.value, int):
            return f"{self.value}"
        return f"{self.value}"



"""RF node implementation."""

import datetime
import logging
from typing import Callable, Dict, List

from .client import AsyncAiriosModbusClient
from .constants import BatteryStatus, FaultStatus, ProductId, RFCommStatus, RFStats
from .data_model import AiriosNodeData
from .exceptions import AiriosException
from .registers import (
    DateRegister,
    DateTimeRegister,
    FloatRegister,
    RegisterAccess,
    RegisterAddress,
    RegisterBase,
    Result,
    StringRegister,
    U16Register,
    U32Register,
)

_LOGGER = logging.getLogger(__name__)


class Reg(RegisterAddress):
    """Register set for RF nodes."""

    RF_ADDRESS = 40000
    PRODUCT_ID = 40002
    SOFTWARE_VERSION = 40004
    OEM_NUMBER = 40005
    RF_CAPABILITIES = 40006
    MANUFACTURE_DATE = 40007
    SOFTWARE_BUILD_DATE = 40009
    PRODUCT_NAME = 40011
    RECEIVED_PRODUCT_ID = 40021
    RF_COMM_STATUS = 40101
    BATTERY_STATUS = 40102
    FAULT_STATUS = 40103
    RF_STATS_INDEX = 40120
    RF_STATS_LENGTH = 40121
    RF_STATS_DEVICE = 40122
    RF_STATS_AVERAGE = 40124
    RF_STATS_STDDEV = 40125
    RF_STATS_MIN = 40127
    RF_STATS_MAX = 40128
    RF_STATS_MISSED = 40129
    RF_STATS_RECEIVED = 40130
    RF_STATS_AGE = 40131
    FAULT_HISTORY_INDEX = 40300
    FAULT_HISTORY_LENGTH = 40301
    FAULT_HISTORY_TIMESTAMP = 40302
    FAULT_HISTORY_FAULTCODE = 40304
    FAULT_HISTORY_STATUS_INFO = 40305
    FAULT_HISTORY_COMM_STATUS = 40307


class AiriosNode:
    """Represents a RF node."""

    client: AsyncAiriosModbusClient
    slave_id: int
    registers: List[RegisterBase] = []
    regmap: Dict[RegisterAddress, RegisterBase] = {}

    def __init__(self, slave_id: int, client: AsyncAiriosModbusClient) -> None:
        """Initialize the node class instance."""
        self.client = client
        self.slave_id = int(slave_id)
        node_registers: List[RegisterBase] = [
            U32Register(Reg.RF_ADDRESS, RegisterAccess.READ),
            U32Register(Reg.PRODUCT_ID, RegisterAccess.READ),
            U16Register(Reg.SOFTWARE_VERSION, RegisterAccess.READ),
            U16Register(Reg.OEM_NUMBER, RegisterAccess.READ),
            U16Register(Reg.RF_CAPABILITIES, RegisterAccess.READ),
            DateRegister(Reg.MANUFACTURE_DATE, RegisterAccess.READ),
            DateRegister(Reg.SOFTWARE_BUILD_DATE, RegisterAccess.READ),
            StringRegister(Reg.PRODUCT_NAME, 10, RegisterAccess.READ),
            U32Register(Reg.RECEIVED_PRODUCT_ID, RegisterAccess.READ),
            U16Register(Reg.RF_COMM_STATUS, RegisterAccess.READ),
            U16Register(Reg.BATTERY_STATUS, RegisterAccess.READ),
            U16Register(Reg.FAULT_STATUS, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_INDEX, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.RF_STATS_LENGTH, RegisterAccess.READ),
            U32Register(Reg.RF_STATS_DEVICE, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_AVERAGE, RegisterAccess.READ),
            FloatRegister(Reg.RF_STATS_STDDEV, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_MIN, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_MAX, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_MISSED, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_RECEIVED, RegisterAccess.READ),
            U16Register(Reg.RF_STATS_AGE, RegisterAccess.READ),
            U16Register(Reg.FAULT_HISTORY_INDEX, RegisterAccess.READ | RegisterAccess.WRITE),
            U16Register(Reg.FAULT_HISTORY_LENGTH, RegisterAccess.READ | RegisterAccess.WRITE),
            DateTimeRegister(Reg.FAULT_HISTORY_TIMESTAMP, RegisterAccess.READ),
            U16Register(Reg.FAULT_HISTORY_FAULTCODE, RegisterAccess.READ),
            U32Register(Reg.FAULT_HISTORY_STATUS_INFO, RegisterAccess.READ),
            U16Register(Reg.FAULT_HISTORY_COMM_STATUS, RegisterAccess.READ),
        ]
        self._add_registers(node_registers)

    def _add_registers(self, reglist: List[RegisterBase]):
        self.registers.extend(reglist)
        self.regmap: Dict[RegisterAddress, RegisterBase] = {
            regdesc.description.address: regdesc for regdesc in self.registers
        }

    async def node_rf_address(self) -> Result[int]:
        """Get the node RF address, also used as node serial number."""
        return await self.client.get_register(self.regmap[Reg.RF_ADDRESS], self.slave_id)

    async def node_product_id(self) -> Result[ProductId]:
        """Get the node product ID.

        This is the value assigned to the virtual node instance created by the bridge when
        a device is bound. The actual received product ID from the real RF node can is
        available in the RECEIVED_PRODUCT_ID register.
        """
        result = await self.client.get_register(self.regmap[Reg.PRODUCT_ID], self.slave_id)
        return Result(ProductId(result.value), None)

    async def node_software_version(self) -> Result[int]:
        """Get the node software version."""
        return await self.client.get_register(self.regmap[Reg.SOFTWARE_VERSION], self.slave_id)

    async def node_oem_number(self) -> Result[int]:
        """Get the node OEM number.

        It is 0x00 or 0xFF when not used.
        """
        return await self.client.get_register(self.regmap[Reg.OEM_NUMBER], self.slave_id)

    async def node_rf_capabilities(self) -> Result[int]:
        """Get the node RF capabilities.

        The value depends on the specific device.
        """
        return await self.client.get_register(self.regmap[Reg.RF_CAPABILITIES], self.slave_id)

    async def node_manufacture_date(self) -> Result[datetime.date]:
        """Get the node manufacture date."""
        return await self.client.get_register(self.regmap[Reg.MANUFACTURE_DATE], self.slave_id)

    async def node_software_build_date(self) -> Result[datetime.date]:
        """Get the node software build date."""
        return await self.client.get_register(self.regmap[Reg.SOFTWARE_BUILD_DATE], self.slave_id)

    async def node_product_name(self) -> Result[str]:
        """Get the node product name."""
        return await self.client.get_register(self.regmap[Reg.PRODUCT_NAME], self.slave_id)

    async def node_received_product_id(self) -> Result[ProductId]:
        """Get the received product ID.

        This is the value received from the bound node. If it does not match register
        NODE_PRODUCT_ID a wrong product is bound.
        """
        result = await self.client.get_register(self.regmap[Reg.RECEIVED_PRODUCT_ID], self.slave_id)
        return Result(ProductId(result.value), result.status)

    async def node_rf_comm_status(self) -> Result[RFCommStatus]:
        """Get the node RF communication status."""
        return await self.client.get_register(self.regmap[Reg.RF_COMM_STATUS], self.slave_id)

    async def node_battery_status(self) -> Result[BatteryStatus]:
        """Get the node battery status."""
        regdesc = self.regmap[Reg.BATTERY_STATUS]
        result = await self.client.get_register(regdesc, self.slave_id)
        available = result.value != 0xFFFF
        low = False
        if available:
            low = result.value != 0
        status = BatteryStatus(available, low)
        return Result(status, result.status)

    async def node_fault_status(self) -> Result[FaultStatus]:
        """Get the node fault status."""
        result = await self.client.get_register(self.regmap[Reg.FAULT_STATUS], self.slave_id)
        available = result.value != 0xFFFF
        fault = result.value != 0
        return Result(FaultStatus(available=available, fault=fault), result.status)

    async def node_clear_rf_stats(self) -> bool:
        """Clears the node RF stats."""
        return await self.client.set_register(self.regmap[Reg.RF_STATS_INDEX], 255, self.slave_id)

    async def node_rf_stats(self) -> RFStats:
        """Get the node RF stats."""
        r = await self.client.get_register(self.regmap[Reg.RF_STATS_LENGTH], self.slave_id)
        nrecs = r.value
        recs: list[RFStats.Record] = []
        for i in range(0, nrecs):
            ok = await self.client.set_register(self.regmap[Reg.RF_STATS_INDEX], i, self.slave_id)
            if not ok:
                _LOGGER.warning("Failed to write %d to RF stats index register", i)
                continue
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_DEVICE], self.slave_id)
            device_id: int = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_AVERAGE], self.slave_id)
            averate: int = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_STDDEV], self.slave_id)
            stddev: float = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_MIN], self.slave_id)
            minimum: int = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_MAX], self.slave_id)
            maximum: int = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_MISSED], self.slave_id)
            missed: int = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_RECEIVED], self.slave_id)
            received: int = r.value
            r = await self.client.get_register(self.regmap[Reg.RF_STATS_AGE], self.slave_id)
            age = datetime.timedelta(minutes=r.value)
            rec = RFStats.Record(
                device_id=device_id,
                averate=averate,
                stddev=stddev,
                minimum=minimum,
                maximum=maximum,
                missed=missed,
                received=received,
                age=age,
            )
            recs.append(rec)
        return RFStats(records=recs)

    async def _safe_fetch(self, fetcher: Callable):
        try:
            result = await fetcher()
        except AiriosException:
            return None
        return result

    async def fetch_node(self) -> AiriosNodeData:  # pylint: disable=duplicate-code
        """Fetch relevant node data at once."""

        return AiriosNodeData(
            slave_id=self.slave_id,
            rf_address=await self._safe_fetch(self.node_rf_address),
            product_id=await self._safe_fetch(self.node_product_id),
            sw_version=await self._safe_fetch(self.node_software_version),
            product_name=await self._safe_fetch(self.node_product_name),
            rf_comm_status=await self._safe_fetch(self.node_rf_comm_status),
            battery_status=await self._safe_fetch(self.node_battery_status),
            fault_status=await self._safe_fetch(self.node_fault_status),
        )



"""RF device implementation."""

from __future__ import annotations

from typing import List

from .client import AsyncAiriosModbusClient
from .constants import BoundStatus, ValueErrorStatus
from .data_model import AiriosDeviceData
from .node import AiriosNode
from .registers import (
    I16Register,
    RegisterAccess,
    RegisterAddress,
    RegisterBase,
    Result,
    U16Register,
)


class Reg(RegisterAddress):
    """Register set for RF devices."""

    RF_LAST_SEEN = 40100
    VALUE_ERROR_STATUS = 40104
    RF_LAST_RSSI = 40109
    BOUND_STATUS = 40110


class AiriosDevice(AiriosNode):
    """Represents a RF device."""

    def __init__(self, slave_id: int, client: AsyncAiriosModbusClient) -> None:
        """Initialize the device class instance."""
        super().__init__(slave_id, client)
        dev_registers: List[RegisterBase] = [
            U16Register(Reg.RF_LAST_SEEN, RegisterAccess.READ),
            U16Register(Reg.VALUE_ERROR_STATUS, RegisterAccess.READ),
            I16Register(Reg.RF_LAST_RSSI, RegisterAccess.READ),
            U16Register(Reg.BOUND_STATUS, RegisterAccess.READ),
        ]
        self._add_registers(dev_registers)

    async def device_bound_status(self) -> Result[BoundStatus]:
        """Get the device bound status."""
        result = await self.client.get_register(self.regmap[Reg.BOUND_STATUS], self.slave_id)
        return Result(BoundStatus(result.value), result.status)

    async def device_value_error_status(self) -> Result[ValueErrorStatus]:
        """Get the device value error status."""
        result = await self.client.get_register(self.regmap[Reg.VALUE_ERROR_STATUS], self.slave_id)
        return Result(ValueErrorStatus(result.value), result.status)

    async def fetch_device(self) -> AiriosDeviceData:  # pylint: disable=duplicate-code
        """Fetch the device data."""

        return AiriosDeviceData(
            slave_id=self.slave_id,
            rf_address=await self._safe_fetch(self.node_rf_address),
            product_id=await self._safe_fetch(self.node_product_id),
            sw_version=await self._safe_fetch(self.node_software_version),
            product_name=await self._safe_fetch(self.node_product_name),
            rf_comm_status=await self._safe_fetch(self.node_rf_comm_status),
            battery_status=await self._safe_fetch(self.node_battery_status),
            fault_status=await self._safe_fetch(self.node_fault_status),
            bound_status=await self._safe_fetch(self.device_bound_status),
            value_error_status=await self._safe_fetch(self.device_value_error_status),
        )



"""Exceptions from the Airios library."""


class AiriosException(Exception):
    """Base class for Airios exceptions."""


class AiriosDecodeError(AiriosException):
    """Decoding failed."""


class AiriosEncodeError(AiriosException):
    """Encoding failed."""


class AiriosConnectionException(AiriosException):
    """Exception connecting to device."""


class AiriosInvalidArgumentException(AiriosException):
    """Invalid argument."""

    def __init__(self, message: str):
        super().__init__(message)


class AiriosReadException(AiriosException):
    """Exception reading register from device."""

    def __init__(self, message: str, modbus_exception_code: int | None):
        super().__init__(message)
        self.modbus_exception_code = modbus_exception_code


class AiriosConnectionInterruptedException(AiriosException):
    """Connection to the device was interrupted."""


class AiriosSlaveBusyException(AiriosException):
    """Non-fatal exception while trying to read from device."""


class AiriosSlaveFailureException(AiriosException):
    """Possibly fatal exception while trying to read from device."""


class AiriosAcknowledgeException(AiriosException):
    """Device accepted the request but needs time to process it."""


class AiriosWriteException(AiriosException):
    """Exception writing register to device."""

    def __init__(self, message: str, modbus_exception_code: int | None):
        super().__init__(message)
        self.modbus_exception_code = modbus_exception_code


class AiriosBindingException(AiriosException):
    """Binding failed."""


class AiriosNotImplemented(AiriosException):
    """Exception not implemented"""


class AiriosIOException(AiriosException):
    """I/O exception"""



"""Airios VMN-05LM02 remote implementation."""

from __future__ import annotations

from typing import List

from .client import AsyncAiriosModbusClient
from .constants import VMDRequestedVentilationSpeed
from .data_model import VMN05LM02Data
from .device import AiriosDevice
from .registers import (
    RegisterAccess,
    RegisterAddress,
    RegisterBase,
    Result,
    U16Register,
)


class Reg(RegisterAddress):
    """Register set for VMN-05LM02 remote node."""

    REQUESTED_VENTILATION_SPEED = 41000


class VMN05LM02(AiriosDevice):
    """Represents a VMN-05LM02 remote node."""

    def __init__(self, slave_id: int, client: AsyncAiriosModbusClient) -> None:
        """Initialize the VMN-05LM02 node instance."""
        super().__init__(slave_id, client)
        vmn_registers: List[RegisterBase] = [
            U16Register(
                Reg.REQUESTED_VENTILATION_SPEED, RegisterAccess.READ | RegisterAccess.STATUS
            ),
        ]
        self._add_registers(vmn_registers)

    def __str__(self) -> str:
        return f"VMN-05LM02@{self.slave_id}"

    async def requested_ventilation_speed(self) -> Result[VMDRequestedVentilationSpeed]:
        """Get the requested ventilation speed."""
        regdesc = self.regmap[Reg.REQUESTED_VENTILATION_SPEED]
        result = await self.client.get_register(regdesc, self.slave_id)
        return Result(VMDRequestedVentilationSpeed(result.value), result.status)

    async def fetch_vmn_data(self) -> VMN05LM02Data:  # pylint: disable=duplicate-code
        """Get the node device data at once."""

        return VMN05LM02Data(
            slave_id=self.slave_id,
            rf_address=await self._safe_fetch(self.node_rf_address),
            product_id=await self._safe_fetch(self.node_product_id),
            sw_version=await self._safe_fetch(self.node_software_version),
            product_name=await self._safe_fetch(self.node_product_name),
            rf_comm_status=await self._safe_fetch(self.node_rf_comm_status),
            battery_status=await self._safe_fetch(self.node_battery_status),
            fault_status=await self._safe_fetch(self.node_fault_status),
            bound_status=await self._safe_fetch(self.device_bound_status),
            value_error_status=await self._safe_fetch(self.device_value_error_status),
            requested_ventilation_speed=await self._safe_fetch(self.requested_ventilation_speed),
        )
